import numpy as _np
import random as _random
from typing import (
    Callable as _Callable,
    List as _List,
    Optional as _Optional,
    Tuple as _Tuple
)

from mykit.kit.math import relu as _ReLU


class DenseNN:
    """Also known as fully connected neural networks"""

    def __init__(
        self,
        sizes: _List[int],
        hidden_act_fn: _Callable[[_np.ndarray, bool], _np.ndarray] = _ReLU,
        output_act_fn: _Callable[[_np.ndarray, bool], _np.ndarray] = _ReLU,
        load_weights_and_biases: _Optional[_Tuple[_List[_np.ndarray], _List[_np.ndarray]]] = None,
    ) -> None:
        """
        The network uses normalized Xavier weight initialization and cross-entropy loss function.

        ---

        `load_weights_and_biases`: Pre-loaded weights and biases decoded by `decode` method. `None` for a new network.
        """

        self.sizes = sizes
        self.hidden_act_fn = hidden_act_fn
        self.output_act_fn = output_act_fn

        if load_weights_and_biases is None:
            self.weights = [
                -_np.sqrt(6/(x+y)) + _np.random.rand(y, x)*2*_np.sqrt(6/(x+y))  # normalized Xavier
                for x, y in zip(self.sizes[:-1], self.sizes[1:])
            ]
            self.biases = [
                _np.random.randn(y, 1)
                for y in self.sizes[1:]
            ]
        else:
            self.weights, self.biases = load_weights_and_biases

        self.n_layer = len(sizes)
        self.n_input = sizes[0]
        self.hidden_layers = sizes[1:-1]
        self.n_hidden_layer = len(self.hidden_layers)
        self.n_output = sizes[-1]

        ## values of these list are from the last feedforwarding
        self.z_values = []
        self.a_values = []
        self.decision = None  # the index of the output neuron with highest value

    def feedforward(self, a: _np.ndarray) -> int:
        """
        feedforward for given input `a`.

        ---

        ## Params
        `a`: Input array with shape `(n, 1)`.
            Example: `np.array([[i1], [i2], [i3], ...])`
        
        ## Returns
        The index of the maximum value in the output array.
        """

        ## reset
        self.z_values = []
        self.a_values = [a]

        ## if `self.n_layer` is 4 -> the variable `idx` below will be: 0, 1, 2
        for idx, (w, b) in enumerate(zip(self.weights, self.biases)):

            z = _np.dot(w, a) + b

            if idx == (self.n_layer - 2):  # output layer
                a = self.output_act_fn(z)
            else:
                a = self.hidden_act_fn(z)

            ## used for backpropagation
            self.z_values.append(z)
            self.a_values.append(a)

        self.decision = a.argmax()  # the index of the maximum value in `a`
        return self.decision

    def _backprop(self, inputs: _np.ndarray, outputs: _np.ndarray):
        """
        backpropagation algorithm.

        ---

        ## Params
        `inputs`: Input array with shape `(n, 1)`.
            Example: `np.array([[i1], [i2], [i3], ...])`
        `outputs`: Desired output array with shape `(m, 1)`.
            Example: `np.array([[o1], [o2], [o3], ...])`

        ## Returns
        `Tuple[dgw, dgb]`: Gradients of weights and biases as lists.
        """

        self.feedforward(inputs)

        ## gradient delta
        dgw = [_np.zeros(w.shape) for w in self.weights]
        dgb = [_np.zeros(b.shape) for b in self.biases]

        ## output layer
        delta = self.a_values[-1] - outputs  # using cross-entropy loss function
        dgw[-1] = _np.dot(delta, self.a_values[-2].transpose())
        dgb[-1] = delta

        ## hidden layers
        for l in range(2, self.n_layer):
            delta = _np.dot(self.weights[-l + 1].transpose(), delta)*self.hidden_act_fn(self.z_values[-l], derivative=True)
            dgw[-l] = _np.dot(delta, self.a_values[-l - 1].transpose())
            dgb[-l] = delta

        return (dgw, dgb)

    def _tuning(self, samples: _List[_Tuple[_np.ndarray, _np.ndarray]], k1: float, k2: float) -> None:
        """
        Update weights and biases.

        ---
        
        ## Params
            - `samples`: A list of input-output pairs used for updating the model
            - For optimization purposes, these values are precalculated:
                - `k1`: `learning_rate*(regularization/n_training_data)`
                - `k2`: `learning_rate/len(samples)`
        """

        ## Initialize gradients
        gw = [_np.zeros(w.shape) for w in self.weights]
        gb = [_np.zeros(b.shape) for b in self.biases]

        ## Compute gradients
        for inputs, outputs in samples:
            dgw, dgb = self._backprop(inputs, outputs)

            gw = [_gw + _dgw for _gw, _dgw in zip(gw, dgw)]
            gb = [_gb + _dgb for _gb, _dgb in zip(gb, dgb)]

        ## Update weights and biases
        self.weights = [
            (1 - k1)*w - k2*_gw
            for w, _gw in zip(self.weights, gw)
        ]
        self.biases = [
            b - k2*_gb
            for b, _gb in zip(self.biases, gb)
        ]

    def train(
        self,
        training_data: _List[_Tuple[_np.ndarray, _np.ndarray]],
        sample_size: int,
        n_epoch: int,
        learning_rate: float = 0.005,
        regularization: float = 0.001
    ) -> None:
        """
        Trains the model using gradient descent with regularization.

        ---

        ## Params
            - `training_data`: A list of tuples containing input-output pairs
            - `sample_size`: The number of samples to use for each iteration of gradient descent
                - `sample_size = 1` for stochastic gradient descent
                - `1 < sample_size < n_training_data` for mini-batch gradient descent
                - `sample_size = n_training_data` for batch gradient descent
            - `n_epoch`: The number of times to iterate over the entire training dataset
            - `learning_rate`: The learning rate used for gradient descent
            - `regularization`: The regularization factor to prevent overfitting
        """

        n_training_data = len(training_data)

        k1 = learning_rate*(regularization/n_training_data)

        for _ in range(n_epoch):

            _random.shuffle(training_data)
            list_of_samples = [
                training_data[i : i + sample_size]
                for i in range(0, n_training_data, sample_size)
            ]

            for samples in list_of_samples:
                self._tuning(samples, k1, learning_rate/len(samples))


    def encode(self) -> dict:
        """Encode the weights and biases as a JSON-serializable dictionary."""
        encoded_data = {
            'weights': [w.tolist() for w in self.weights],
            'biases': [b.tolist() for b in self.biases]
        }
        return encoded_data

    @staticmethod
    def decode(encoded_data: dict) -> _Tuple[_List[_np.ndarray], _List[_np.ndarray]]:
        """Decode the weights and biases from what is encoded by the `encode` method."""
        weights = [_np.array(w) for w in encoded_data['weights']]
        biases = [_np.array(b) for b in encoded_data['biases']]
        return weights, biases


import math as _math
import random as _random
import tkinter as _tk
from typing import (
    Dict as _Dict,
    List as _List,
    Optional as _Optional,
    Tuple as _Tuple,
    Union as _Union
)

from mykit.kit.math import (
    get_angle as _get_angle,
    rotate as _rotate
)


class Arrow:

    ## <runtime>

    _page: _tk.Canvas = None
    @staticmethod
    def _set_page(page):
        Arrow._page = page

    arrows: _Dict[str, 'Arrow'] = {}
    arrow_tags: _Dict[str, _List['Arrow']] = {}

    ## </runtime>

    def __init__(
        self,
        from_x: int, from_y: int,
        to_x: int, to_y: int,
        /,
        color: str = '#ddd',
        width_rod: int = 1,
        width_tip: int = 2,
        tip_len: float = 20,
        tip_angle: float = 35,
        visible: bool = True,
        
        id: _Optional[str] = None,
        tags: _Optional[_Union[str, _List[str]]] = None,
    ) -> None:

        ## <dependencies check>

        ## reminder: Arrow isn't actually a widget. it can be used as a utility
        ##           outside App's purposes. so just use Arrow._set_page to achieve it.
        if Arrow._page is None:
            raise AssertionError('Can\'t use widgets before App initialized.')
        
        ## </dependencies check>
        
        self.from_x = from_x
        self.from_y = from_y
        self.to_x = to_x
        self.to_y = to_y

        self.color = color
        self.width_rod = width_rod
        self.width_tip = width_tip
        self.tip_len = tip_len
        self.tip_angle = tip_angle

        self.visible = visible

        ## <id>

        ## to make sure that we can modify a specific arrow without affecting the others
        if id is None:
            self.id = _random.randint(-10000, 10000)
            while self.id in Arrow.arrows:
                self.id = _random.randint(-10000, 10000)
        else:
            self.id = id
            if self.id in Arrow.arrows:
                raise ValueError(f'The Arrow\'s id {repr(id)} is duplicated.')
        
        Arrow.arrows[self.id] = self

        ## </id>

        ## <tags>

        if type(tags) is str:
            self.tags = [tags]
        elif (type(tags) is list) or (type(tags) is tuple) or (tags is None):
            self.tags = tags
        
        if tags is not None:
            for tag in self.tags:
                if tag in Arrow.arrow_tags:
                    Arrow.arrow_tags[tag].append(self)
                else:
                    Arrow.arrow_tags[tag] = [self]
        
        ## </tags>

        ## init
        self._redraw()

    def _redraw(self):

        Arrow._page.delete(f'Arrow_{self.id}')

        if self.visible:
            Arrow._page.create_line(
                self.from_x, self.from_y,
                self.to_x, self.to_y,
                fill=self.color, width=self.width_rod, tags=f'Arrow_{self.id}'
            )

            ## <creating the tip>
            ## for `angle`: remember to flip the y-sign because tkinter's y-positive direction towards the bottom
            angle = _get_angle(self.from_x, -self.from_y, self.to_x, -self.to_y)
            
            tipx = self.tip_len*_math.sin(self.tip_angle*_math.pi/180)
            tipy = self.tip_len*_math.cos(self.tip_angle*_math.pi/180)
            
            ## Remember `tip_left` and `tip_right` with y-positive towards the top,
            ## as they transformed under normal Cartesian coordinates.
            tip_left = _rotate(-tipx, -tipy, 0, 0, angle)
            tip_right = _rotate(tipx, -tipy, 0, 0, angle)

            ## Revert to the tkinter coordinate scheme, where y-positive is oriented downwards.
            tip_left = (self.to_x+tip_left[0], self.to_y-tip_left[1])
            tip_right = (self.to_x+tip_right[0], self.to_y-tip_right[1])

            tip_points = [tip_left, (self.to_x, self.to_y), tip_right]
            Arrow._page.create_line(tip_points, fill=self.color, width=self.width_tip, tags=f'Arrow_{self.id}')
            ## </creating the tip>


    def set_visibility(self, visible: bool, /):
        if visible != self.visible:
            self.visible = visible
            self._redraw()

    @staticmethod
    def set_visibility_by_id(id: str, visible: bool, /):
        Arrow.arrows[id].set_visibility(visible)

    @staticmethod
    def set_visibility_by_tag(tag: str, visible: bool, /):
        for arrow in Arrow.arrow_tags[tag]:
            arrow.set_visibility(visible)

    @staticmethod
    def set_visibility_all(visible: bool, /):
        for arrow in Arrow.arrows.values():
            arrow.set_visibility(visible)


#!/usr/bin/env python
import argparse
import os
import sys

try:
    from mykit import __version__, LIB_NAME
except ModuleNotFoundError:
    ## This is done to handle situations where the 'mykit' isn't installed yet, so it can't be accessed.
    root_pth = os.path.dirname(os.path.dirname(__file__))
    sys.path.append(root_pth)
    from mykit import __version__, LIB_NAME


def main():

    parser = argparse.ArgumentParser(prog=LIB_NAME)
    parser.add_argument('-v', '--version', action='version', version=f'%(prog)s-{__version__}')

    args = parser.parse_args()  # run the parser


if __name__ == '__main__':
    main()


import os as _os
import uuid as _uuid


## Ref: https://github.com/orgs/community/discussions/28146

def set_output(name, value):
    """
    ## Demo
    >>> set_output('foo', '123')
    """
    with open(_os.environ['GITHUB_OUTPUT'], 'a') as fh:
        print(f'{name}={value}', file=fh)

def set_multiline_output(name, value):
    with open(_os.environ['GITHUB_OUTPUT'], 'a') as fh:
        delimiter = _uuid.uuid1()
        print(f'{name}<<{delimiter}', file=fh)
        print(value, file=fh)
        print(delimiter, file=fh)


import numpy as _np
from typing import (
    Tuple as _Tuple,
    Union as _Union
)


def tanh(x: _Union[float, _np.ndarray], /, derivative: bool = False) -> _Union[float, _np.ndarray]:

    z = _np.tanh(x)

    if derivative:
        return 1 - z*z
    else:
        return z

def sigmoid(x: _Union[float, _np.ndarray], /, derivative: bool = False) -> _Union[float, _np.ndarray]:

    z = 1 / (1 + _np.exp(-x))

    if derivative:
        return z*(1 - z)
    else:
        return z

def relu(x: _Union[float, _np.ndarray], /, derivative: bool = False) -> _Union[float, _np.ndarray]:

    if derivative:
        return _np.where(x > 0, 1, 0)
    else:
        return _np.maximum(0, x)


def get_angle(origin_x: float, origin_y: float, x: float, y: float, in_radians: bool = True) -> float:
    """
    Calculate the angle between two points (origin_x, origin_y) and (x, y) relative to the vertical axis.
    The x-axis is positive to the right, and the y-axis is positive upwards.

    ---

    ## Params
    - `in_radians`: Determines whether the angle is returned in radians (default) or degrees.
        
    ## Returns
    - Angle relative to vertical axis (CCW)

    ## Demo
    >>> print(get_angle(0, 0, 0, 1, False))
    0.0
    >>> print(get_angle(0, 0, -1, 0, False))
    90.0
    >>> print(get_angle(0, 0, 0, -1, False))
    180.0
    >>> print(get_angle(0, 0, 1, 0 , False))
    270.0
    >>> print(get_angle(0, 0, -1, 1 , False))
    45.0
    """

    dx = x - origin_x
    dy = y - origin_y

    ## Quadrant I (top left)
    if (dx <= 0) and (dy > 0):
        angle = _np.arctan2(-dx, dy)
    
    ## Quadrant II (down left)
    elif (dx < 0) and (dy <= 0):
        angle = _np.pi/2 + _np.arctan2(-dy, -dx)

    ## Quadrant III (down right)
    elif (dx >= 0) and (dy < 0):
        angle = _np.pi + _np.arctan2(dx, -dy)
    
    ## Quadrant IV (down right)
    elif (dx > 0) and (dy >= 0):
        angle = _np.pi*3/2 + _np.arctan2(dy, dx)

    if not in_radians:
        angle = _np.degrees(angle)

    return angle


def rotate(x: float, y: float, ctrx: float, ctry: float, a: float) -> _Tuple[float, float]:
    """
    Rotate points `(x, y)` around a center point `(ctrx, ctry)` by a given angle `a` in radians (CCW).
    The x-axis is positive to the right, and the y-axis is positive upwards.
    """

    ## translate the coordinate to the origin
    translated_x = x - ctrx
    translated_y = y - ctry

    ## apply the rotation transformation
    rotated_x = translated_x*_np.cos(a) - translated_y*_np.sin(a)
    rotated_y = translated_x*_np.sin(a) + translated_y*_np.cos(a)

    ## translate the coordinate back to its original position
    final_x = rotated_x + ctrx
    final_y = rotated_y + ctry

    return final_x, final_y


import re as _re
import subprocess as _sp
from typing import (
    List as _List,
    Tuple as _Tuple,
    Union as _Union
)


def get_resolution(input_path: str, ffprobe_path: str) -> _Tuple[int, int]:
    """
    image/video resolution.

    `input_path`: absolute path to the input file
    `ffprobe_path`: absolute path to ffprobe executable

    return: `(width, height)`
    """
    cmd = [
        ffprobe_path,
        '-v', 'error',
        '-select_streams', 'v',
        '-of', 'csv=p=0',
        '-show_entries', 'stream=width,height',
        input_path
    ]
    res = _sp.run(cmd, capture_output=True, text=True)
    match = _re.search(r'^(\d+),(\d+)$', res.stdout.strip())
    width, height = int(match.group(1)), int(match.group(2))
    return width, height


def get_audio_sample_rate(input_path: str, ffprobe_path: str) -> float:
    res = _sp.check_output(
        [
            ffprobe_path, '-v', 'error',
            '-select_streams', 'a',
            '-of', 'csv=p=0',
            '-show_entries', 'stream=sample_rate',
            input_path
        ],
        stderr=_sp.STDOUT, text=True
    )
    reg = _re.match(r'^(?P<v>\d+(?:\.\d+)?)', res)
    return float(reg.group('v'))


def get_vid_dur(file: str, ffprobe: str, /) -> float:
    """
    Get video duration in secs.
    Note, video and audio might be different.
    `file`: full path to the input file
    `ffprobe`: ffprobe.exe full path
    """
    stdout = _sp.check_output(
        [
            ffprobe, '-v', 'error',
            '-select_streams', 'v',
            '-of', 'csv=p=0',
            '-show_entries', 'stream=duration',
            file
        ],
        stderr=_sp.STDOUT, text=True
    )
    res = _re.match(r'^(?P<dur>\d+(?:\.\d+)?)', stdout)
    return float(res.group('dur'))

def get_audio_dur(file: str, ffprobe: str, /) -> float:
    """
    Get audio duration in secs.
    Note, video and audio might be different.
    `file`: full path to the input file
    `ffprobe`: ffprobe.exe full path
    """
    stdout = _sp.check_output(
        [
            ffprobe, '-v', 'error',
            '-select_streams', 'a',
            '-of', 'csv=p=0',
            '-show_entries', 'stream=duration',
            file
        ],
        stderr=_sp.STDOUT, text=True
    )
    res = _re.match(r'^(?P<dur>\d+(?:\.\d+)?)', stdout)
    return float(res.group('dur'))


def get_vid_fps(file: str, ffprobe: str, /, *, do_round: bool = False) -> _Union[int, float]:
    """
    Video fps (average frame rate).

    `file`: full path to the input file
    `ffprobe`: ffprobe.exe full path
    `do_round`: round the fps to the nearest integer
    """
    stdout = _sp.check_output(
        [
            ffprobe, '-v', 'error',
            '-select_streams', 'v',
            '-of', 'csv=p=0',
            '-show_entries', 'stream=avg_frame_rate',
            file
        ],
        stderr=_sp.STDOUT, text=True
    )
    res = _re.match(r'^(?P<fps>\d+/\d+)', stdout)
    fps = eval(res.group('fps'), {})
    if do_round:
        return round(fps)
    else:
        return fps


def gen_dyn_vol(__timestamp: _List[_Tuple[float, float]], /, precision: float = 0.0001, vol_round: int = 2) -> str:
    """
    Generate a dynamic volume (with linear interpolation) filter.

    ---

    ## params
    - `precision`: Lower values indicate better quality with fewer artifacts.
    - `vol_round`: To minimize output length.

    ## Demo
    - If `__timestamp = [(0, 1), (5, 0.5), (10, 1)]` -> Start at full volume,
        transition to half volume, reach half volume at the 5-sec mark,
        transition back to full volume, and return back to full volume at the 10-sec mark.
    - If `__timestamp = [(0, 1), (5, 0.5), (8, 0.5), (10, 1)]` -> Start at full volume,
        transition to half volume, reach half volume at the 5-sec mark, remains at half volume until 8-sec mark,
        transition back to full volume, and return back to full volume at the 10-sec mark.

    ## TODO:
    - Add argument for supporting different interpolation types
    """

    ts_round = len(str(abs(precision)).split('.')[1]) if '.' in str(precision) else 0

    keys = []
    for idx, (current_ts, current_vol) in enumerate(__timestamp):

        if idx == 0:
            V = round(current_vol, vol_round)
            T1 = round(current_ts, ts_round)
            T2 = round(current_ts + precision, ts_round)  # This line is necessary to handle potential occurrences of '1 + 0.1 = 1.100000000000003'.
            keys.append(f'{V}*between(t,{T1},{T2})')
        else:
            prev_ts, prev_vol = __timestamp[idx - 1]
            if idx == 1:
                V1 = round(prev_vol, vol_round)
                V2 = round(prev_vol - current_vol, vol_round)
                T_START = round(prev_ts + precision, ts_round)
                INTERVAL = round(current_ts - prev_ts - 2*precision, ts_round)
                T1 = round(prev_ts + 2*precision, ts_round)
                T2 = round(current_ts, ts_round)
                keys.append(
                    f'({V1}-{V2}*(t-{T_START})/{INTERVAL})'
                    f'*between(t,{T1},{T2})'
                )
            else:
                V1 = round(prev_vol, vol_round)
                V2 = round(prev_vol - current_vol, vol_round)
                T_START = round(prev_ts + precision, ts_round)
                INTERVAL = round(current_ts - prev_ts - precision, ts_round)
                T1 = round(prev_ts + precision, ts_round)
                T2 = round(current_ts, ts_round)
                keys.append(
                    f'({V1}-{V2}*(t-{T_START})/{INTERVAL})'
                    f'*between(t,{T1},{T2})'
                )

    keys_str = '+'.join(keys)
    return f"volume='{keys_str}':eval=frame"


import random as _random
import tkinter as _tk
from typing import (
    Callable as _Callable,
    Dict as _Dict,
    List as _List,
    Literal as _Literal,
    Optional as _Optional,
    Tuple as _Tuple,
    Union as _Union
)


## reminder: Label is different because it uses tkinter.Label
##           instead of tkinter.Canvas. Not sure if it'll be okay,
##           but for now, let's go with it.  @June 18, 2023 - Nicholas
class Label:

    labels: _Dict[str, 'Label'] = {}
    label_tags: _Dict[str, _List['Label']] = {}

    def __init__(
        self,
        x: int = 0,
        y: int = 0,
        text: str = '',
        font: _Union[str, _Tuple[str, int]] = 'Verdana 10',
        justify: str = 'left',
        anchor: _Literal['center', 'n', 'ne', 'e', 'se', 's', 'sw', 'w', 'nw'] = 'nw',
        fg: str = '#ccc',
        bg: str = '#111',
        bd: str = '#555',
        bd_width: int = 0,
        wraplength: int = 1000,
        padx: int = 0,
        pady: int = 0,
        visible: bool = True,

        id: _Optional[str] = None,
        tags: _Optional[_Union[str, _List[str]]] = None,
    ):
        """
        `x`: the x-position for the `anchor` argument, not the center of the label
        `y`: the y-position for the `anchor` argument, not the center of the label
        """

        self.x = x  # reminder: this is the x-position for the `anchor` argument, not the center of the label
        self.y = y  # reminder: this is the y-position for the `anchor` argument, not the center of the label
        self.font = font
        self.text = text
        self.anchor = anchor
        self.fg = fg
        self.visible = visible

        self.default_text = text

        self.label = _tk.Label(
            text=text, font=font, justify=justify,
            foreground=fg, background=bg,
            highlightbackground=bd, highlightthickness=bd_width,
            wraplength=wraplength,
            padx=padx, pady=pady
        )

        if visible:
            self.label.place(x=x, y=y, anchor=anchor)


        ## <id>

        ## self.id ensures that we can modify a specific instance without affecting the others
        if id is None:
            self.id = str(_random.randint(0, 100_000))
            while self.id in Label.labels:
                self.id = str(_random.randint(0, 100_000))
        else:
            self.id = id
            if self.id in Label.labels:
                raise ValueError(f'The id {repr(id)} is duplicated.')
        
        Label.labels[self.id] = self

        ## </id>


        ## <tags>
        
        if type(tags) is str:
            self.tags = [tags]
        elif (type(tags) is list) or (type(tags) is tuple) or (tags is None):
            self.tags = tags
        
        if tags is not None:
            for tag in self.tags:
                if tag in Label.label_tags:
                    Label.label_tags[tag].append(self)
                else:
                    Label.label_tags[tag] = [self]
        
        ## <tags>


    def set_text(self, text: _Optional[str], /):
        """if None -> set default text."""

        if text is None:
            text = self.default_text

        if text != self.text:
            self.text = text
            self.label.configure(text=text)

    @staticmethod
    def set_text_by_id(id: str, text: _Optional[str], /):
        Label.labels[id].set_text(text)


    def set_font(self, font: _Union[str, _Tuple[str, int]], /):
        if font != self.font:
            self.font = font
            self.label.configure(font=font)

    @staticmethod
    def set_font_by_id(id: str, font: _Union[str, _Tuple[str, int]], /):
        Label.labels[id].set_font(font)


    def set_fg(self, fg: str, /):
        if fg != self.fg:
            self.fg = fg
            self.label.configure(fg=fg)

    @staticmethod
    def set_fg_by_id(id: str, fg: str, /):
        Label.labels[id].set_fg(fg)


    def set_visibility(self, visible: bool, /):
        if self.visible is not visible:
            self.visible = visible
            if visible:
                self.label.place(x=self.x, y=self.y, anchor=self.anchor)
            else:
                self.label.place_forget()

    @staticmethod
    def set_visibility_by_id(id: str, visible: bool, /):
        Label.labels[id].set_visibility(visible)

    @staticmethod
    def set_visibility_by_tag(tag: str, visible: bool, /):
        for label in Label.label_tags[tag]:
            label.set_visibility(visible)

    @staticmethod
    def set_visibility_all(visible: bool, /):
        for label in Label.labels.values():
            label.set_visibility(visible)


    def get_width(self) -> int:
        return self.label.winfo_reqwidth()

    @staticmethod
    def get_width_by_id(id: str, /) -> int:
        return Label.labels[id].get_width()

    def get_height(self) -> int:
        return self.label.winfo_reqheight()

    @staticmethod
    def get_height_by_id(id: str, /) -> int:
        return Label.labels[id].get_height()


    def get_anchor_loc(self, anchor: _Literal['center', 'n', 'ne', 'e', 'se', 's', 'sw', 'w', 'nw'], /) -> _Tuple[int, int]:
        """To get the center of the label coordinates, use `anchor='center'`"""

        W = self.get_width()
        H = self.get_height()

        ## get the coordinates for the center (X, Y)
        if self.anchor == 'center':
            X = self.x
            Y = self.y
        elif self.anchor == 'n':
            X = self.x
            Y = self.y + H/2
        elif self.anchor == 'ne':
            X = self.x - W/2
            Y = self.y + H/2
        elif self.anchor == 'e':
            X = self.x - W/2
            Y = self.y
        elif self.anchor == 'se':
            X = self.x - W/2
            Y = self.y - H/2
        elif self.anchor == 's':
            X = self.x
            Y = self.y - H/2
        elif self.anchor == 'sw':
            X = self.x + W/2
            Y = self.y - H/2
        elif self.anchor == 'w':
            X = self.x + W/2
            Y = self.y
        elif self.anchor == 'nw':
            X = self.x + W/2
            Y = self.y + H/2

        ## returning the requested anchor location
        if anchor == 'center':
            return (X, Y)
        elif anchor == 'n':
            return (X, Y-H/2)
        elif anchor == 'ne':
            return (X+W/2, Y-H/2)
        elif anchor == 'e':
            return (X+W/2, Y)
        elif anchor == 'se':
            return (X+W/2, Y+H/2)
        elif anchor == 's':
            return (X, Y+H/2)
        elif anchor == 'sw':
            return (X-W/2, Y+H/2)
        elif anchor == 'w':
            return (X-W/2, Y)
        elif anchor == 'nw':
            return (X-W/2, Y-H/2)
        else:
            raise ValueError(f'Invalid anchor value: {repr(anchor)}')

    @staticmethod
    def get_anchor_loc_by_id(id: str, anchor: _Literal['center', 'n', 'ne', 'e', 'se', 's', 'sw', 'w', 'nw'], /) -> _Tuple[int, int]:
        return Label.labels[id].get_anchor_loc(anchor)

    
    def move(
        self,
        x: int,
        y: int,
        /,
        anchor: _Optional[_Literal['center', 'n', 'ne', 'e', 'se', 's', 'sw', 'w', 'nw']] = None
    ) -> 'Label':
        """If `anchor = None`, the current anchor will be used."""
        self.x = x
        self.y = y
        if anchor is not None:
            self.anchor = anchor
        # self.label.place(x=x, y=y, anchor=anchor)  # reminder: don't do this because `anchor` could be `None`
        self.label.place(x=x, y=y, anchor=self.anchor)
        return self

    @staticmethod
    def move_by_id(
        id: str,
        x: int,
        y: int,
        /,
        anchor: _Optional[_Literal['center', 'n', 'ne', 'e', 'se', 's', 'sw', 'w', 'nw']] = None
    ) -> None:
        Label.labels[id].move(x, y, anchor)


    def align(
        self,
        target: 'Label',
        anchor: str = 'nw',
        target_anchor: str = 'ne',
        xgap: float = 15,
        ygap: float = 0
    ) -> 'Label':
        """
        Valid options for `anchor` and `target_anchor` are `['center', 'n', 'ne', 'e', 'se', 's', 'sw', 'w', 'nw']`.
        """

        ## getting the target anchor location
        x, y = target.get_anchor_loc(target_anchor)

        ## shifting
        x += xgap
        y += ygap

        ## moving the label
        self.move(x, y, anchor)

        ## return the instance so that this method can also be used when
        ## creating the instance, like `label_2 = Label(text='foo').align(label_1)`.
        return self


    def destroy(self) -> None:
        Label.labels.pop(self.id)

        if self.tags is not None:
            for tag in list(self.tags):
                Label.label_tags[tag].remove(self)
                if Label.label_tags[tag] == []:
                    Label.label_tags.pop(tag)
        
        self.label.destroy()
    
    @staticmethod
    def destroy_by_id(id: str, /) -> None:
        Label.labels[id].destroy()

    @staticmethod
    def destroy_by_tag(tag: str, /) -> None:
        for label in list(Label.label_tags[tag]):
            label.destroy()

    @staticmethod
    def destroy_all() -> None:
        for label in list(Label.labels.values()):
            label.destroy()


from typing import (
    Callable as _Callable
)


class Eventdriven:
    """Simple event-driven implementation"""

    bus = {}  # Container

    def register(name: str, /) -> None:
        if name in Eventdriven.bus:
            raise ValueError(f'Event {repr(name)} already exists.')
        Eventdriven.bus[name] = []

    def call(name: str, /) -> None:
        for fn in Eventdriven.bus[name]:
            fn()

    def listen(to: str, do: _Callable[[], None]) -> None:
        if to not in Eventdriven.bus:
            raise ValueError(f'Event {repr(to)} does not exist.')
        Eventdriven.bus[to].append(do)


import subprocess as _sp

from mykit.kit.color import (
    Hex as _Hex,
    Colored as _Colored
)


def _logger(level, color, msg):
    _sp.run(['echo', "[$(date +'%H:%M:%S')] " + _Colored(level, color) + ': ' + msg])


class eL:
    """
    eL (`echo` Log): A simple logger using `echo` function,
    intended for use within GitHub Action virtual machines.
    Inspired by `mykit.kit.pLog.pL`.
    """

    def group(name:str, /) -> None:
        _sp.run(['echo', f'::group::{name}'])

    def endgroup() -> None:
        _sp.run(['echo', '::endgroup::'])

    def debug(msg:str, /) -> None:
        _logger('DEBUG', _Hex.WHEAT, msg)

    def info(msg:str, /) -> None:
        _logger('INFO', _Hex.BLUE_GRAY, msg)

    def warning(msg:str, /) -> None:
        _logger('WARNING', _Hex.DARK_ORANGE, msg)

    def error(msg:str, /) -> None:
        _logger('ERROR', _Hex.SCARLET, msg)


import re as _re
import os as _os
from typing import (
    Any as _Any,
    Dict as _Dict,
    List as _List,
    NoReturn as _NoReturn,
    Tuple as _Tuple,
    Union as _Union
)

from mykit.kit.path import open_file as _open_file
from mykit.kit.utils import is_valid_var_name as _is_valid_var_name


class KeyCrate:
    """
    Using a .txt file (which is easy to open, read, and modify)
    to store key-value pairs, aiming to keep it simple and fast.
    """

    def __init__(
        self,
        file_pth: str,
        /,
        key_is_var: bool = False,
        eval_value: bool = False,
        only_keys: _Union[_List[str], _Tuple[str, ...], None] = None,
        need_keys: _Union[_List[str], _Tuple[str, ...], None] = None
    ) -> None:
        """
        Storing key-value pairs (key: value) in the .txt file `file_pth`.
        Access the value using either the class-like way (`kc.key`)
        or the dictionary-like way (`kc['key']`).

        ---

        ## Params
        - `file_pth`: full path to the .txt file that holds the key-value pairs
        - `key_is_var`: if `True`, all keys must be valid as variable name
        - `eval_value`: if `True`, all values must be evaluatable by `eval()`
        - `only_keys`: only keys specified by `only_keys` are allowed in `file_pth`
        - `need_keys`: keys that specified by `need_keys` must exist in `file_pth`

        ## Exceptions
        - `ValueError`: if `file_pth` is not a .txt file
        - `FileNotFoundError`: if `file_pth` is not a file
        - `AttributeError`: if trying to access a nonexistent key
        - see `self.parse` exceptions

        ## Demo 1
        >>> kc = KeyCrate('config.txt', eval_value=True)
        >>> fps = kc.fps  # or kc['fps']
        >>> output_dir = kc['output folder']

        ## Demo 2
        >>> settings = KeyCrate('settings.txt', key_is_var=True, only_keys=['fps', 'dur'])
        >>> fps = settings.fps
        >>> dur = settings.dur
        """

        ## keycrate file should be a .txt file
        if not file_pth.endswith('.txt'):
            raise ValueError(f'KeyCrate file {repr(file_pth)} should be a .txt file.')

        ## keycrate file must exist
        if not _os.path.isfile(file_pth):
            raise FileNotFoundError(f'KeyCrate file {repr(file_pth)} is not found.')


        ## added the prefix "_kc__" to prevent conflicts with the keys

        self._kc__file_pth = file_pth
        self._kc__key_is_var = key_is_var
        self._kc__eval_value = eval_value

        self._kc__only_keys = only_keys
        if type(only_keys) is tuple:
            self._kc__only_keys = list(only_keys)

        self._kc__need_keys = need_keys
        if type(need_keys) is tuple:
            self._kc__need_keys = list(need_keys)


        ## init
        self.parse()

    def __getattr__(self, __name: str, __default: _Any = None, /) -> _NoReturn:
        """This method is called when attribute `__name` is not found"""
        raise AttributeError(f'KeyCrate file {repr(self._kc__file_pth)} does not have key {repr(__name)}.')

    def __getitem__(self, __key: str, /) -> _Any:
        """
        To access the keys in a dictionary-like way (e.g., `kc['key1']`),
        it is commonly used for keys that are not variable names (e.g., `kc['full name']`).
        """
        return getattr(self, __key)

    def _read(self) -> str:
        """Return the current content of the file."""
        with open(self._kc__file_pth, 'r') as fp:
            out = fp.read()
        return out

    def parse(self) -> None:
        """
        If necessary, rerun this function to reparse.

        ---

        ## Exceptions
        - `SyntaxError`: if invalid syntax is found
        - `ValueError`: if any duplicate keys are found
        - `AssertionError`: if key is invalid for variable name (when `key_is_var=True`)
        - `AssertionError`: if value can't be evaluated (when `eval_value=True`)
        - `AssertionError`: if unexpected key is found (only when `only_keys` is specified)
        - `AssertionError`: if missing key is found (only when `need_keys` is specified)
        """

        raw = self._read()

        if self._kc__need_keys is not None:
            _need_keys = self._kc__need_keys.copy()

        for line_no, line in enumerate(raw.split('\n'), 1):

            if (line == '') or _re.match(r'\s*#--.*', line):
                ## blank-line or comment-header
                continue

            res = _re.match(
                r'\s*(?P<key>.+?)\s*:\s*(?P<val>.+?)\s*(?:#--.*)?;',
                ## added ";" to capture "everything" in `val` (but not
                ## the *last-spaces and comment* after it). The pattern doesn't
                ## work without it
                line + ';'
            )
            if res is None:
                raise SyntaxError(
                    f'KeyCrate file {repr(self._kc__file_pth)} '
                    f'has invalid syntax at line {line_no}: {repr(line)}'
                )

            key = res.group('key')
            val = res.group('val')

            if key in self.__dict__:
                raise ValueError(
                    f'KeyCrate file {repr(self._kc__file_pth)} '
                    f'has a duplicated key {repr(key)} found at line {line_no}.'
                )

            if self._kc__key_is_var:
                if not _is_valid_var_name(key):
                    raise AssertionError(
                        f'KeyCrate file {repr(self._kc__file_pth)} '
                        f'has a key {repr(key)} that is invalid for a variable name, found at line {line_no}.'
                    )

            if self._kc__eval_value:
                try:
                    val = eval(val, {})
                except (NameError, SyntaxError):
                    raise AssertionError(
                        f'KeyCrate file {repr(self._kc__file_pth)} '
                        f'has a value {repr(val)} that cannot be evaluated, found at line {line_no}.'
                    )

            if self._kc__only_keys is not None:
                if key not in self._kc__only_keys:
                    raise AssertionError(
                        f'KeyCrate file {repr(self._kc__file_pth)} '
                        f'has an unexpected key {repr(key)} found at line {line_no}.'
                    )

            if self._kc__need_keys:
                try:
                    _need_keys.remove(key)
                except ValueError:
                    ## note that if `need_keys` is specified, other keys
                    ## (not in `need_keys`) are also allowed.
                    ## so, this exception must be handled
                    pass

            setattr(self, key, val)

        if self._kc__need_keys:
            if _need_keys != []:
                raise AssertionError(
                    f'KeyCrate file {repr(self._kc__file_pth)} is missing keys: '
                    + ', '.join(map(repr, _need_keys))
                )

    def open(self) -> None:
        """open the .txt file"""
        _open_file(self._kc__file_pth)
    
    def export(self) -> _Dict[str, _Any]:
        """Return all the key-value pairs as dictionary."""

        force_copy = eval(str(self.__dict__), {})  # not sure gonna copy all the things deeply

        out = {
            key: force_copy[key]
            for key in filter(lambda key: not key.startswith('_kc__'), force_copy.keys())
        }
        return out


import random as _random
import tkinter as _tk
from typing import (
    Dict as _Dict,
    List as _List,
    Optional as _Optional,
    Tuple as _Tuple,
    Union as _Union
)


class Biplot:

    ## <runtime>

    _page: _tk.Canvas = None
    @staticmethod
    def _set_page(page):
        Biplot._page = page

    biplots: _Dict[str, 'Biplot'] = {}
    biplot_tags: _Dict[str, _List['Biplot']] = {}

    ## </runtime>

    def __init__(
        self,
        points1: _List[_Tuple[float, float]] = [],
        points2: _List[_Tuple[float, float]] = [],

        xmin: _Optional[float] = None,
        xmax: _Optional[float] = None,
        ymin: _Optional[float] = None,
        ymax: _Optional[float] = None,

        width: int = 300,
        height: int = 200,
        tl_x: int = 0,
        tl_y: int = 0,

        pad_x: float = 0.03,
        pad_y: float = 0.07,
        show_tick: bool = True,
        show_grid: bool = True,
        ntick_x: int = 10,
        ntick_y: int = 10,
        tick_len: int = 12,
        arrow_size: int = 7,
        arrow_width: int = 2,

        grid_color: str = '#555',
        axes_color: str = '#ccc',
        axes_label_color: str = '#ccc',

        title: str = '',
        title_color: str = '#fff',
        title_font: _Union[str, tuple] = ('Arial Bold', 15),

        x_axis_label: str = '',
        x_axis_label_shift: int = 15,
        x_axis_label_font: _Union[str, tuple] = ('Arial Bold', 12),
        y_axis_label: str = '',
        y_axis_label_shift: int = 15,
        y_axis_label_font: _Union[str, tuple] = ('Arial Bold', 12),

        tick_x_prefix: str = '',
        tick_x_suffix: str = '',
        tick_x_shift: int = 0,
        tick_x_font: _Union[str, tuple] = 'Consolas 9',
        tick_x_prec: int = 1,
        tick_y_prefix: str = '',
        tick_y_suffix: str = '',
        tick_y_shift: int = 0,
        tick_y_font: _Union[str, tuple] = 'Consolas 9',
        tick_y_prec: int = 1,
        tick_color: str = '#ccc',

        plot_color1: str = '#7f7',
        plot_color2: str = '#f77',
        plot_thick: int = 1,  # line width

        show_points: bool = False,
        points_rad: int = 7,
        points_color: str = '#a77',
        points_border: str = '#eee',

        legend1: _Optional[str] = None,
        legend2: _Optional[str] = None,
        legends_bar_width: int = 30,
        legends_bar_height: int = 6,
        legends_pad_x: int = 7,
        legends_pad_y: int = 20,
        legends_font: _Union[str, tuple] = ('Arial Bold', 13),
        legends_color: str = '#f5f5f5',
        legends_shift_x: int = 30,
        legends_shift_y: int = 10,

        visible: bool = True,

        id: _Optional[str] = None,
        tags: _Optional[_Union[str, _List[str]]] = None,
    ):
        """
        Please ensure that `points1` and `points2` always have the same length.
        The graph can be displayed without any given points,
        but in order to show the plot, two points need to be specified.

        ---
        
        ## Params
        - `xrange`: if `None` -> using the x-range from `points`
        - `yrange`: if `None` -> using the y-range from `points`
        """

        ## <dependencies check>

        ## reminder: Biplot isn't actually a widget. it can be used as a utility
        ##           outside App's purposes. so just use Biplot._set_page to achieve it.
        if Biplot._page is None:
            raise AssertionError('Can\'t use widgets before App initialized.')
        
        ## </dependencies check>

        self.points1 = points1
        self.points2 = points2

        self.xmin = xmin
        self.xmax = xmax
        self.ymin = ymin
        self.ymax = ymax

        self.width = width
        self.height = height
        self.tl_x = tl_x
        self.tl_y = tl_y

        self.pad_x = pad_x
        self.pad_y = pad_y
        self.show_tick = show_tick
        self.show_grid = show_grid
        self.ntick_x = ntick_x
        self.ntick_y = ntick_y
        self.tick_len = tick_len
        self.arrow_size = arrow_size
        self.arrow_width = arrow_width

        self.grid_color = grid_color
        self.axes_color = axes_color
        self.axes_label_color = axes_label_color

        self.title = title
        self.title_color = title_color
        self.title_font = title_font

        self.x_axis_label = x_axis_label
        self.x_axis_label_shift = x_axis_label_shift
        self.x_axis_label_font = x_axis_label_font
        self.y_axis_label = y_axis_label
        self.y_axis_label_shift = y_axis_label_shift
        self.y_axis_label_font = y_axis_label_font

        self.tick_x_prefix = tick_x_prefix
        self.tick_x_suffix = tick_x_suffix
        self.tick_x_shift = tick_x_shift
        self.tick_x_font = tick_x_font
        self.tick_x_prec = tick_x_prec
        self.tick_y_prefix = tick_y_prefix
        self.tick_y_suffix = tick_y_suffix
        self.tick_y_shift = tick_y_shift
        self.tick_y_font = tick_y_font
        self.tick_y_prec = tick_y_prec
        self.tick_color = tick_color

        self.plot_color1 = plot_color1
        self.plot_color2 = plot_color2
        self.plot_thick = plot_thick

        self.show_points = show_points
        self.points_rad = points_rad
        self.points_color = points_color
        self.points_border = points_border

        self.legend1 = legend1
        self.legend2 = legend2
        self.legends_bar_width = legends_bar_width
        self.legends_bar_height = legends_bar_height
        self.legends_pad_x = legends_pad_x
        self.legends_pad_y = legends_pad_y
        self.legends_font = legends_font
        self.legends_color = legends_color
        self.legends_shift_x = legends_shift_x
        self.legends_shift_y = legends_shift_y

        self.visible = visible

        ## <id>

        ## `self.id`: to make sure that we can modify a specific instance without affecting the others
        if id is None:
            self.id = _random.randint(-10000, 10000)
            while self.id in Biplot.biplots:
                self.id = _random.randint(-10000, 10000)
        else:
            self.id = id
            if self.id in Biplot.biplots:
                raise ValueError(f'The id {repr(id)} is duplicated.')
        
        Biplot.biplots[self.id] = self

        ## </id>

        ## <tags>

        if type(tags) is str:
            self.tags = [tags]
        elif (type(tags) is list) or (type(tags) is tuple) or (tags is None):
            self.tags = tags

        if tags is not None:
            for tag in self.tags:
                if tag in Biplot.biplot_tags:
                    Biplot.biplot_tags[tag].append(self)
                else:
                    Biplot.biplot_tags[tag] = [self]
        
        ## </tags>


        ## <preprocessing>
        self.plot_width = width*(1-pad_x)
        self.plot_height = height*(1-pad_y)
        ## </preprocessing>


        ## init
        self._redraw()

    def _redraw(self):
        """redraw the entire graph"""

        ## 2 points need to be specified in order to draw the graph
        if len(self.points1) < 3:
            self.points1 = [(0, 0), (1, 0)]
            self.points2 = [(0, 0), (1, 0)]

        x_values = [p[0] for p in self.points1]
        if self.xmin is None:
            XMIN = min(x_values)
        else:
            XMIN = self.xmin
        if self.xmax is None:
            XMAX = max(x_values)
        else:
            XMAX = self.xmax

        y_values = [p[1] for p in (self.points1 + self.points2)]
        if self.ymin is None:
            YMIN = min(y_values)
        else:
            YMIN = self.ymin
        if self.ymax is None:
            YMAX = max(y_values)
        else:
            YMAX = self.ymax

        LEN_X = XMAX - XMIN
        LEN_Y = YMAX - YMIN


        ## title
        Biplot._page.create_text(
            self.tl_x+self.width/2, self.tl_y,
            text=self.title, font=self.title_font, fill=self.title_color,
            tags=f'Biplot_{self.id}'
        )


        ## grids
        if self.show_grid:

            ## vertical grids
            for x in range(self.ntick_x):
                X = self.tl_x + ((x+1)/self.ntick_x)*self.plot_width
                Biplot._page.create_line(
                    X, self.tl_y+(self.height-self.plot_height),
                    X, self.tl_y+self.height,
                    fill=self.grid_color, width=1, tags=f'Biplot_{self.id}'
                )

            ## horizontal grids
            for y in range(self.ntick_y):
                Y = self.tl_y+self.height - ((y+1)/self.ntick_y)*self.plot_height
                Biplot._page.create_line(
                    self.tl_x                , Y,
                    self.tl_x+self.plot_width, Y,
                    fill=self.grid_color, width=1, tags=f'Biplot_{self.id}'
                )


        ## x-axis
        Biplot._page.create_line(
            self.tl_x           , self.tl_y+self.height,
            self.tl_x+self.width, self.tl_y+self.height,
            fill=self.axes_color, width=1, tags=f'Biplot_{self.id}'
        )
        ## x-axis arrow
        Biplot._page.create_line(
            self.tl_x+self.width-self.arrow_size, self.tl_y+self.height-self.arrow_size,
            self.tl_x+self.width                , self.tl_y+self.height,
            self.tl_x+self.width-self.arrow_size, self.tl_y+self.height+self.arrow_size,
            fill=self.axes_color, width=self.arrow_width, tags=f'Biplot_{self.id}'
        )
        ## x-axis label
        Biplot._page.create_text(
            self.tl_x+self.width+self.x_axis_label_shift, self.tl_y+self.height,
            text=self.x_axis_label, anchor='w', fill=self.axes_label_color, font=self.x_axis_label_font, tags=f'Biplot_{self.id}'
        )

        ## y-axis
        Biplot._page.create_line(
            self.tl_x, self.tl_y,
            self.tl_x, self.tl_y+self.height,
            fill=self.axes_color, width=1, tags=f'Biplot_{self.id}'
        )
        ## y-axis arrow
        Biplot._page.create_line(
            self.tl_x-self.arrow_size, self.tl_y+self.arrow_size,
            self.tl_x, self.tl_y,
            self.tl_x+self.arrow_size, self.tl_y+self.arrow_size,
            fill=self.axes_color, width=self.arrow_width, tags=f'Biplot_{self.id}'
        )
        ## y-axis label
        Biplot._page.create_text(
            self.tl_x, self.tl_y-self.y_axis_label_shift,
            text=self.y_axis_label, anchor='s', fill=self.axes_label_color, font=self.y_axis_label_font, tags=f'Biplot_{self.id}'
        )


        ## ticks
        if self.show_tick:

            ## x-axis ticks
            for x in range(self.ntick_x+1):
                X = self.tl_x + (x/self.ntick_x)*self.plot_width

                ## tick
                Biplot._page.create_line(
                    X, self.tl_y+self.height-self.tick_len/2,
                    X, self.tl_y+self.height+self.tick_len/2,
                    fill=self.tick_color, width=1, tags=f'Biplot_{self.id}'
                )

                ## tick-label
                _num = XMIN + (x/self.ntick_x)*(LEN_X)
                if self.tick_x_prec == 0:
                    _num = int(_num)
                else:
                    _num = round(_num, self.tick_x_prec)
                text = self.tick_x_prefix + str(_num) + self.tick_x_suffix
                Biplot._page.create_text(
                    X, self.tl_y+self.height+self.tick_len+self.tick_x_shift,
                    text=text, anchor='n', font=self.tick_x_font, fill=self.axes_label_color,
                    tags=(f'Biplot_{self.id}', f'Biplot_{self.id}_ticks')
                )

            ## y-axis ticks
            for y in range(self.ntick_y+1):
                Y = self.tl_y+self.height - (y/self.ntick_y)*self.plot_height

                ## tick
                Biplot._page.create_line(
                    self.tl_x-self.tick_len/2, Y,
                    self.tl_x+self.tick_len/2, Y,
                    fill=self.tick_color, width=1, tags=f'Biplot_{self.id}'
                )

                ## tick-label
                _num = YMIN + (y/self.ntick_y)*(LEN_Y)
                if self.tick_y_prec == 0:
                    _num = int(_num)
                else:
                    _num = round(_num, self.tick_y_prec)
                text = self.tick_y_prefix + str(_num) + self.tick_y_suffix
                Biplot._page.create_text(
                    self.tl_x-self.tick_len-self.tick_y_shift, Y,
                    text=text, anchor='e', font=self.tick_y_font, fill=self.axes_label_color,
                    tags=(f'Biplot_{self.id}', f'Biplot_{self.id}_ticks')
                )


        ## plot 1
        coords1 = []
        for x, y in self.points1:
            X = self.tl_x + (x - XMIN)*(self.plot_width/LEN_X)
            Y = self.tl_y + self.height - (y - YMIN)*(self.plot_height/LEN_Y)
            coords1.append((X, Y))
        Biplot._page.create_line(
            coords1,
            fill=self.plot_color1, width=self.plot_thick,
            tags=(f'Biplot_{self.id}', f'Biplot_{self.id}_plot')
        )

        ## plot 2
        coords2 = []
        for x, y in self.points2:
            X = self.tl_x + (x - XMIN)*(self.plot_width/LEN_X)
            Y = self.tl_y + self.height - (y - YMIN)*(self.plot_height/LEN_Y)
            coords2.append((X, Y))
        Biplot._page.create_line(
            coords2,
            fill=self.plot_color2, width=self.plot_thick,
            tags=(f'Biplot_{self.id}', f'Biplot_{self.id}_plot')
        )

        if self.show_points:
            for x, y in (coords1 + coords2):
                Biplot._page.create_oval(
                    x-self.points_rad/2, y-self.points_rad/2,
                    x+self.points_rad/2, y+self.points_rad/2,
                    fill=self.points_color, outline=self.points_border, width=1, tags=f'Biplot_{self.id}'
                )
        

        ## legends
        if self.legend1 is not None:
            Biplot._page.create_rectangle(
                self.tl_x+self.legends_shift_x, self.tl_y+self.legends_shift_y,
                self.tl_x+self.legends_shift_x+self.legends_bar_width, self.tl_y+self.legends_shift_y+self.legends_bar_height,
                fill=self.plot_color1, width=0, tags=f'Biplot_{self.id}'
            )
            Biplot._page.create_text(
                self.tl_x+self.legends_shift_x+self.legends_bar_width+self.legends_pad_x,
                self.tl_y+self.legends_shift_y+self.legends_bar_height/2,
                text=self.legend1, font=self.legends_font, justify='left', anchor='w', fill=self.legends_color,
                tags=f'Biplot_{self.id}'
            )
        if self.legend2 is not None:
            Biplot._page.create_rectangle(
                self.tl_x+self.legends_shift_x,
                self.tl_y+self.legends_shift_y+self.legends_pad_y,
                self.tl_x+self.legends_shift_x+self.legends_bar_width,
                self.tl_y+self.legends_shift_y+self.legends_pad_y+self.legends_bar_height,
                fill=self.plot_color2, width=0, tags=f'Biplot_{self.id}'
            )
            Biplot._page.create_text(
                self.tl_x+self.legends_shift_x+self.legends_bar_width+self.legends_pad_x,
                self.tl_y+self.legends_shift_y+self.legends_pad_y+self.legends_bar_height/2,
                text=self.legend2, font=self.legends_font, justify='left', anchor='w', fill=self.legends_color,
                tags=f'Biplot_{self.id}'
            )

    def redraw_plot(self, points1: _List[_Tuple[float, float]], points2: _List[_Tuple[float, float]], /) -> None:
        """
        Redraws the plot and updates the tick labels with a new set of given points.
        """

        ## reminder: to optimize things, only redraw the necessary part
        ##           so the code below is duplicated from `_redraw`.
        ##           it's redundant, but currently the easiest way to
        ##           achieve the desired functionality

        self.points1 = points1
        self.points2 = points2

        ## 2 points need to be specified in order to draw the graph
        if len(self.points1) < 3:
            self.points1 = [(0, 0), (1, 0)]
            self.points2 = [(0, 0), (1, 0)]

        x_values = [p[0] for p in self.points1]
        if self.xmin is None:
            XMIN = min(x_values)
        else:
            XMIN = self.xmin
        if self.xmax is None:
            XMAX = max(x_values)
        else:
            XMAX = self.xmax

        y_values = [p[1] for p in (self.points1 + self.points2)]
        if self.ymin is None:
            YMIN = min(y_values)
        else:
            YMIN = self.ymin
        if self.ymax is None:
            YMAX = max(y_values)
        else:
            YMAX = self.ymax

        ## LEN_X and LEN_Y can't be zero
        LEN_X = max(10**(-self.tick_x_prec), XMAX - XMIN)
        LEN_Y = max(10**(-self.tick_y_prec), YMAX - YMIN)


        ## redraw the ticks

        Biplot._page.delete(f'Biplot_{self.id}_ticks')

        if self.show_tick:

            ## x-axis ticks
            for x in range(self.ntick_x+1):
                X = self.tl_x + (x/self.ntick_x)*self.plot_width

                ## tick-label
                _num = XMIN + (x/self.ntick_x)*(LEN_X)
                if self.tick_x_prec == 0:
                    _num = int(_num)
                else:
                    _num = round(_num, self.tick_x_prec)
                text = self.tick_x_prefix + str(_num) + self.tick_x_suffix
                Biplot._page.create_text(
                    X, self.tl_y+self.height+self.tick_len+self.tick_x_shift,
                    text=text, anchor='n', font=self.tick_x_font, fill=self.axes_label_color,
                    tags=(f'Biplot_{self.id}', f'Biplot_{self.id}_ticks')
                )

            ## y-axis ticks
            for y in range(self.ntick_y+1):
                Y = self.tl_y+self.height - (y/self.ntick_y)*self.plot_height

                ## tick-label
                _num = YMIN + (y/self.ntick_y)*(LEN_Y)
                if self.tick_y_prec == 0:
                    _num = int(_num)
                else:
                    _num = round(_num, self.tick_y_prec)
                text = self.tick_y_prefix + str(_num) + self.tick_y_suffix
                Biplot._page.create_text(
                    self.tl_x-self.tick_len-self.tick_y_shift, Y,
                    text=text, anchor='e', font=self.tick_y_font, fill=self.axes_label_color,
                    tags=(f'Biplot_{self.id}', f'Biplot_{self.id}_ticks')
                )


        ## redraw the plot

        Biplot._page.delete(f'Biplot_{self.id}_plot')

        ## plot 1
        coords1 = []
        for x, y in self.points1:
            X = self.tl_x + (x - XMIN)*(self.plot_width/LEN_X)
            Y = self.tl_y + self.height - (y - YMIN)*(self.plot_height/LEN_Y)
            coords1.append((X, Y))
        Biplot._page.create_line(
            coords1,
            fill=self.plot_color1, width=self.plot_thick,
            tags=(f'Biplot_{self.id}', f'Biplot_{self.id}_plot')
        )

        ## plot 2
        coords2 = []
        for x, y in self.points2:
            X = self.tl_x + (x - XMIN)*(self.plot_width/LEN_X)
            Y = self.tl_y + self.height - (y - YMIN)*(self.plot_height/LEN_Y)
            coords2.append((X, Y))
        Biplot._page.create_line(
            coords2,
            fill=self.plot_color2, width=self.plot_thick,
            tags=(f'Biplot_{self.id}', f'Biplot_{self.id}_plot')
        )

    def shift_plot(self, new_points1: _List[_Tuple[float, float]], new_points2: _List[_Tuple[float, float]], /) -> None:
        """
        Shifts the plot by inserting `new_points` and removing the leftmost points.
        For example, if 2 pairs are inserted, 2 leftmost pairs will be removed.
        Make sure that the new x values are always greater than the rightmost x value.
        For instance, if the current rightmost x is 100, the new x values
        should follow a sequence like 100, 101, 102.
        """
        
        n_new = len(new_points1)
        
        points1 = self.points1[n_new:] + new_points1
        points2 = self.points2[n_new:] + new_points2
        
        self.redraw_plot(points1, points2)

    def add_point(self, point1: _Tuple[float, float], point2: _Tuple[float, float], max: _Optional[int] = None):
        """
        Adding the new point to the current plot.
        If a max value is specified (e.g., max=100), and the total number of points
        exceeds 100, the plot will shift by inserting the new point and removing
        the leftmost point. If max is not specified, the point will simply be added.
        """

        points1 = self.points1
        points2 = self.points2

        if max is None:
            points1.append(point1)
            points2.append(point2)
            self.redraw_plot(points1, points2)
        else:
            if len(points1) < max:
                points1.append(point1)
                points2.append(point2)
                self.redraw_plot(points1, points2)
            else:
                self.shift_plot([point1], [point2])


    def set_visibility(self, visible: bool, /):
        if self.visible is not visible:
            self.visible = visible
            self._redraw()

    @staticmethod
    def set_visibility_by_id(id: str, visible: bool, /):
        Biplot.biplots[id].set_visibility(visible)

    @staticmethod
    def set_visibility_by_tag(tag: str, visible: bool, /):
        for bp in Biplot.biplot_tags[tag]:
            bp.set_visibility(visible)

    @staticmethod
    def set_visibility_all(visible: bool, /):
        for bp in Biplot.biplots.values():
            bp.set_visibility(visible)


import os as _os
import pkg_resources as _pr


try:
    pkg = _pr.get_distribution('mykit')
except _pr.DistributionNotFound:
    ## This exception happens during GitHub Actions testing, let's handle it this way.
    class _Testing:
        version = 'testing'
        project_name = 'mykit'
    pkg = _Testing


__version__ = pkg.version


LIB_REPO = 'https://github.com/nvfp/mykit'
LIB_NAME = 'mykit'
LIB_DIST_NAME = pkg.project_name  # distribution name

_ROOT_DIR_PTH = _os.path.dirname(_os.path.dirname(_os.path.abspath(__file__)))  # should only be used for development
DIST_DIR_PTH = _os.path.join(_ROOT_DIR_PTH, LIB_DIST_NAME)
# print(f'DEBUG: _ROOT_DIR_PTH = {_ROOT_DIR_PTH}')
# print(f'DEBUG: DIST_DIR_PTH = {DIST_DIR_PTH}')

APP_DIR_PTH = _os.path.join(DIST_DIR_PTH, 'app')
KIT_DIR_PTH = _os.path.join(DIST_DIR_PTH, 'kit')
REC_DIR_PTH = _os.path.join(DIST_DIR_PTH, 'rec')
WEB_DIR_PTH = _os.path.join(DIST_DIR_PTH, 'web')
# print(f'DEBUG: APP_DIR_PTH = {APP_DIR_PTH}')



import random as _random
import numpy as _np
from typing import (
    Callable as _Callable,
    Dict as _Dict,
    List as _List,
    Optional as _Optional,
    Tuple as _Tuple
)

from mykit.kit.neuralnet.dense import DenseNN as _DenseNN


class GeneticNN:

    def __init__(
        self,
        layer_sizes: _List[int],
        hidden_act: _Callable[[_np.ndarray, bool], _np.ndarray],
        output_act: _Callable[[_np.ndarray, bool], _np.ndarray],

        population_size: int,
        crossover_threshold: float = 0.001,
        mutation1_rate: float = 0.85,
        mutation2_rate: float = 0.99,
        mutation2_range: _Tuple[float, float] = (0.1, 2.5),
        n_new: int = 0,
        init_score: float = 0,
        
        load_wnb: _Optional[_Tuple[_List[_np.ndarray], _List[_np.ndarray]]] = None,
    ) -> None:
        """
        Genetic neural network using dense neural network.

        ---

        ## Params
        - `layer_sizes`: The size of the dense neural network for each individual.
        - `population_size`: Number of individuals per generation.
        - `crossover_threshold`: Determines how weights/biases from parent1 and parent2 are
                                 combined during crossover. If the difference between parent1
                                 and parent2 is less than this threshold, the child's value
                                 is the average; otherwise, it's randomly selected from
                                 either parent1 or parent2.
        - `mutation1_rate`: The probability of mutating a weight or bias (parent1/parent2 are close). Range: 0-1.
        - `mutation2_rate`: The probability of mutating a weight or bias (parent1/parent2 are different). Range: 0-1.
        - `mutation2_range`: The range of mutation change. For example, if set to (0.001, 0.1),
                             the weight or bias can be offset by a random value between 0.001
                             and 0.1, or between -0.001 and -0.1.
        - `n_new`: Number of new individuals (neither parent nor child) for each subsequent generation,
                   with the constraint `n_new < population_size-3`.
        - `init_score`: Initial score
        - `load_wnb`: Populate the first generation with pretrained weights and biases (use this
                      if you don't want to start the training from the beginning again).
        """

        self.layer_sizes = layer_sizes
        self.hidden_act = hidden_act
        self.output_act = output_act

        self.population_size = population_size
        self.crossover_threshold = crossover_threshold
        self.mutation1_rate = mutation1_rate
        self.mutation2_rate = mutation2_rate
        self.mutation2_range = mutation2_range
        self.n_new = n_new
        self.init_score = init_score

        self.population: _Dict[_DenseNN, float] = {}
        for _ in range(population_size):

            if load_wnb is None:
                wnb = None
            else:
                w = [w.copy() for w in load_wnb[0]]
                b = [b.copy() for b in load_wnb[1]]
                wnb = (w, b)

            nn = _DenseNN(layer_sizes, hidden_act, output_act, wnb)
            self.population[nn] = init_score


        ## runtime

        self.the_parent: _Tuple[_DenseNN, _DenseNN] = None  # to store the parent that produced the current generation (after executing `keep_elites()`)
        self.prev: _Dict[_DenseNN, float] = {}  # to store the previous generation


    def set_score(self, individual: _DenseNN, score: float, /) -> None:
        self.population[individual] = score

    def get_score(self, individual: _DenseNN, /) -> float:
        return self.population[individual]

    def get_score_min(self) -> float:
        """Return the lowest score of all individuals."""
        return min(self.population.values())

    def get_score_max(self) -> float:
        """Return the highest score of all individuals."""
        return max(self.population.values())

    def get_score_avg(self) -> float:
        """Return the average score of all individuals."""
        return sum(self.population.values()) / self.population_size

    def get_individual_by_rank(self, rank: int) -> _DenseNN:
        """Returns the individual with the given rank: 0 for the best, 1 for the second best, and so on."""
        return sorted(self.population, key=self.population.get, reverse=True)[rank]

    def get_best_individual(self) -> _DenseNN:
        """Returns the best individual (with the highest score) from the current generation."""
        return self.get_individual_by_rank(0)

    def keep_elites(self) -> None:
        """
        The idea is to continually bring out the most fit individuals from one generation to the next.
        Note: Remember to use this function after updating the scores and before executing `next_generation()`.

        ---

        What's the idea behind this function?
        =====================================

        Let's say there are only three individuals in each generation.

        - first generation -

            1. self.population = {A:0, B:0, C:0} <- each score is set to 0

            2. a new environment is set

            3. A B C undergo the scoring system and let's say the scores:
               self.population = {A:2, B:3, C:1}

            4. `keep_elites()`:

               sorted:
               {B:3, A:2, C:1}

               store the parent:
               self.the_parent = (B, A)

               store the prev:
               self.prev = self.population = {A:2, B:3, C:1}

        - 2nd generation -
            
            1. `next_generation()`:
               
               A and B are the top two, producing one child D.
               now, the current population:
               self.population = {A:0, B:0, D:0}

            2. a new environment is set

            3. A B D undergo the scoring system

            == Case I (Overall Upgrade) ==

                3. let's say the scores:
                   self.population = {A:4, B:5, D:7}

                4. `keep_elites()`:

                    picking the fitter parent:
                    comparing `self.prev = {A:2, B:3, C:1}` with `self.population = {A:4, B:5, D:7}`
                    result: self.prev = {C:1} and self.population = {A:4, B:5, D:7}

                    combine:
                    {A:4, B:5, D:7, C:1}

                    sort:
                    {D:7, B:5, A:4, C:1}

                    pick the fittest:
                    {D:7, B:5, A:4}

                    set to current population:
                    self.population = {D:7, B:5, A:4}

                    store the parent:
                    self.the_parent = (D, B)

                    stored in self.prev:
                    self.prev = self.population = {D:7, B:5, A:4}

                In other words, A and B work better in this environment.
                This is the ideal case where overall improvement occurs.

            == Case II (Overall Downgrade) ==
                
                3. let's say the scores:
                   self.population = {A:0, B:0, D:0}

                4. `keep_elites()`:

                    picking the fitter parent:
                    comparing `self.prev = {A:2, B:3, C:1}` with `self.population = {A:0, B:0, D:0}`
                    result: self.prev = {A:2, B:3, C:1} and self.population = {D:0}

                    combine:
                    {D:0, A:2, B:3, C:1}

                    sort:
                    {B:3, A:2, C:1, D:0}

                    pick the fittest:
                    {B:3, A:2, C:1}

                    set to current population:
                    self.population = {B:3, A:2, C:1}

                    ########################################################
                    let's contrast keeping the fittest from the previous generation versus not keeping them

                    if not:
                        current population:
                        self.population = {A:0, B:0, D:0}

                        potential parents: A/B or A/D or B/D
                        A/B is the same as before
                        but A/D or B/D is something different (see below for the considerations)

                    if yes:
                        current population:
                        self.population = {B:3, A:2, C:1}

                        parent: A and B
                    ########################################################
                
                    store the parent:
                    self.the_parent = (B, A)

                    stored in self.prev:
                    self.prev = self.population = {B:3, A:2, C:1}

                Note that while A B C has already experimented with 2 different environments, D has only done so once.
                In other words, the "veterans" are prioritized, meaning we stick with
                the previous generation and move to the next environment, even if A and B
                perform poorly in the current environment.

            Case III (Mix):

                3. let's say the scores:
                   self.population = {A:5, B:0, D:0}
                
                4. `keep_elites()`:

                    picking the fitter parent:
                    comparing `self.prev = {A:2, B:3, C:1}` with `self.population = {A:5, B:0, D:0}`
                    result: self.prev = {B:3, C:1} and self.population = {A:5, D:0}

                    combine:
                    {A:5, D:0, B:3, C:1}

                    sort:
                    {A:5, B:3, C:1, D:0}

                    pick the fittest:
                    {A:5, B:3, C:1}

                    set to current population:
                    self.population = {A:5, B:3, C:1}
                
                    store the parent:
                    self.the_parent = (A, B)

                    stored in self.prev:
                    self.prev = self.population = {A:5, B:3, C:1}

                In other words, stick with the previous generation and move to the next environment.
        """

        ## handle the first generation case
        if self.the_parent is None:
            parent1, parent2 = sorted(self.population, key=self.population.get, reverse=True)[:2]
            self.the_parent = (parent1, parent2)
            self.prev = self.population
        else:

            ## picking the fitter parent
            for parent in self.the_parent:
                old = self.prev[parent]
                new = self.population[parent]
                if new >= old:
                    self.prev.pop(parent)
                else:
                    self.population.pop(parent)

            ## combine
            combined = self.population
            combined.update(self.prev)

            ## sort
            sorted_ = dict(sorted(combined.items(), key=lambda item: item[1], reverse=True))

            ## pick the fittest (pick the top `self.population_size`)
            fittest = {k: sorted_[k] for k in list(sorted_)[:self.population_size]}

            ## set to current population
            self.population = fittest

            parent1, parent2 = sorted(self.population, key=self.population.get, reverse=True)[:2]
            self.the_parent = (parent1, parent2)
            self.prev = self.population

    def next_generation(self) -> None:
        """
        When executed, this function updates `self.population` by selecting the top 2 individuals
        with the highest scores from the current `self.population` and applying crossover
        and mutation to produce the new generation (the new `self.population`).

        Note: Remember to execute this after executing `keep_elites()`.

        Reminder:
        - weight and bias may have different mutation methods
        """

        ## getting the best two parents with the highest scores
        parent1, parent2 = self.the_parent

        self.population = {parent1: self.init_score, parent2: self.init_score}

        ## reproduce
        for _ in range(self.population_size-2-self.n_new):

            child = _DenseNN(self.layer_sizes, self.hidden_act, self.output_act)

            ## weights
            for l in range(len(self.layer_sizes)):  # iterate through each layer
                for c in range(self.layer_sizes[l]):  # iterate through each neuron in the current layer

                    if l != (len(self.layer_sizes)-1):
                        for s in range(self.layer_sizes[l+1]):  # iterate through each neuron in the subsequent layer
                            
                            w1 = parent1.weights[l][s, c]
                            w2 = parent2.weights[l][s, c]

                            ## crossover
                            if abs(w2-w1) < self.crossover_threshold:
                                w = (w1+w2)/2
                                ## mutation-1
                                if _random.random() < self.mutation1_rate:
                                    w = _np.random.randn()
                            else:
                                w = _random.choice([w1, w2])
                                ## mutation-2
                                if _random.random() < self.mutation2_rate:
                                    w += _random.uniform(*self.mutation2_range)*_random.choice([-1, 1])
                                    if abs(w) > 5:  # prevent explosion
                                        w = _np.random.randn()

                            child.weights[l][s, c] = w
                    
            ## biases
            for i in range(len(self.layer_sizes)-1):  # note that the input layer does not have biases
                for j in range(self.layer_sizes[i+1]):
                    
                    b1 = parent1.biases[i][j, 0]
                    b2 = parent2.biases[i][j, 0]

                    ## crossover
                    if abs(b2-b1) < self.crossover_threshold:
                        b = (b1+b2)/2
                        ## mutation-1
                        if _random.random() < self.mutation1_rate:
                            b = _np.random.randn()
                    else:
                        b = _random.choice([b1, b2])
                        ## mutation-2
                        if _random.random() < self.mutation2_rate:
                            b += _random.uniform(*self.mutation2_range)*_random.choice([-1, 1])
                            if abs(b) > 5:
                                b = _np.random.randn()

                    child.biases[i][j, 0] = b

            ## add it to the population
            self.population[child] = self.init_score

        for _ in range(self.n_new):
            self.population[_DenseNN(self.layer_sizes, self.hidden_act, self.output_act)] = self.init_score



import os as _os
import sys as _sys
from typing import (
    Tuple as _Tuple
)


def interpolate_color(color1: str, color2: str, x: float) -> str:
    """
    Interpolates between two colors based on the given ratio `x`.

    ---

    ## Params
    - `color1`: The first color in hexadecimal format (e.g., '#RRGGBB').
    - `color2`: The second color in hexadecimal format (e.g., '#RRGGBB').
    - `x`: The ratio determining the interpolation between the two colors. Should be between 0 and 1.

    ## Returns
    - The interpolated color as a hexadecimal string.

    ## Demo
    >>> interpolate_color('#ff0000', '#0000ff', 0.0)
    '#ff0000'
    >>> interpolate_color('#ff0000', '#0000ff', 0.5)
    '#7f007f'
    >>> interpolate_color('#ff0000', '#0000ff', 1.0)
    '#0000ff'
    """
    ## convert color strings to RGB values
    r1, g1, b1 = int(color1[1:3], 16), int(color1[3:5], 16), int(color1[5:7], 16)
    r2, g2, b2 = int(color2[1:3], 16), int(color2[3:5], 16), int(color2[5:7], 16)

    ## interpolate RGB values based on x
    r = int(r1 + (r2 - r1)*x)
    g = int(g1 + (g2 - g1)*x)
    b = int(b1 + (b2 - b1)*x)

    ## convert interpolated RGB values to hexadecimal color string
    interpolated_color = f'#{r:02x}{g:02x}{b:02x}'
    return interpolated_color


def getgray(alpha: float, /, max_lum: int = 255) -> str:
    """
    Returns a hexadecimal color value representing a grayscale shade based on the given alpha and maximum luminance.

    ---

    ## Params
    - `alpha`: A grayscale shade intensity value in the range [0, 1].
    - `max_lum`: Maximum luminance value for grayscale in the range [0, 255].

    ## Demo
    >>> getgray(0.5)
    '#808080'
    """
    a = f'{round(max_lum*alpha):02x}'
    return f'#{a}{a}{a}'


def hex_to_rgb(hex_color: str, /) -> _Tuple[int, int, int]:
    hex_color = hex_color.lstrip('#')
    return tuple(int(hex_color[i:i+2], 16) for i in (0, 2, 4))

def rgb_to_hex(r: int, g: int, b: int, /) -> str:
    return f'#{r:02x}{g:02x}{b:02x}'


def hexa_to_hex(foreground: str, opacity: float, background: str) -> str:
    """
    Calculates the hexadecimal color code of `foreground` with the given `opacity` on `background`.
    The `foreground` and `background` must be valid hexadecimal color codes,
    and the `opacity` value must be in the interval [0, 1].
    """

    fg = [int(foreground[i:i+2], 16) for i in (1, 3, 5)]
    bg = [int(background[i:i+2], 16) for i in (1, 3, 5)]

    r = round(fg[0]*opacity + bg[0]*(1 - opacity))
    g = round(fg[1]*opacity + bg[1]*(1 - opacity))
    b = round(fg[2]*opacity + bg[2]*(1 - opacity))

    return f'#{r:02x}{g:02x}{b:02x}'


def interpolate_with_black(foreground: str, opacity: float) -> str:
    """
    This is the optimized version of `hexa_to_hex(foreground, opacity, '#000000')`.
    Please refer to the documentation of the `hexa_to_hex` function for more details.
    """

    c = [int(foreground[i:i+2], 16) for i in (1, 3, 5)]

    r = round( c[0]*opacity )
    g = round( c[1]*opacity )
    b = round( c[2]*opacity )

    return f'#{r:02x}{g:02x}{b:02x}'


class Hex:
    """List of colors in hexadecimal format."""

    ## Basic

    BLACK = '#000000'
    WHITE = '#ffffff'
    RED = '#ff0000'
    LIME = '#00ff00'
    BLUE = '#0000ff'
    YELLOW = '#ffff00'
    CYAN = '#00ffff'
    MAGENTA = '#ff00ff'
    SILVER = '#c0c0c0'
    GRAY = '#808080'
    MAROON = '#800000'
    OLIVE = '#808000'
    GREEN = '#008000'
    PURPLE = '#800080'
    TEAL = '#008080'
    NAVY = '#000080'

    ## Source: https://en.wikipedia.org/wiki/List_of_colors:_A%E2%80%93F

    ALABASTER = '#edebe0'
    AMBER = '#ffc000'
    ARYLIDE_YELLOW = '#ead76c'
    ASH_GRAY = '#b3c0b6'
    ATOMIC_TANGERINE = '#ff9966'
    AUBURN = '#a62929'
    AUREOLIN = '#fdee00'
    AZURE = '#0080ff'

    BEIGE = '#f5f5dc'
    BITTERSWEET = '#ff715f'
    BLACK_OLIVE = '#3b3d36'
    BLEU_DE_FRANCE = '#318de9'
    BLIZZARD_BLUE = '#ace6ee'
    BLUE_GRAY = '#669acd'
    BLUSH = '#df5c83'
    BOTTLE_GREEN = '#006c4f'
    BRIGHT_MAROON = '#c22147'
    BURGUNDY = '#800021'
    BURNT_ORANGE = '#cc5400'
    BURNT_SIENNA = '#e87352'

    CADMIUM_ORANGE = '#ed872e'
    CAMEL = '#c2996b'
    CAPRI = '#00c0ff'
    CARDINAL = '#c41f3b'
    CARIBBEAN_GREEN = '#00cd9a'
    CAROLINA_BLUE = '#57a1d4'
    CHAMPAGNE = '#f7e8cf'
    CITRINE = '#e4d20a'
    CORAL = '#ff7e4f'
    CORN = '#faed5c'

    DARK_ORANGE = '#ff8c00'
    DARK_RED = '#8c0000'
    DIM_GRAY = '#696969'

    ELECTRIC_BLUE = '#7dfaff'
    EMERALD = '#4fc778'

    FALU_RED = '#801717'
    FLAME = '#e45a21'
    FIREBRICK = '#b32121'

    GO_GREEN = '#00ab66'
    GRANITE_GRAY = '#666666'

    HARLEQUIN = '#40ff00'

    IVORY = '#fffff1'

    JADE = '#00a96c'

    LAPIS_LAZULI = '#26619c'
    LAVENDER_GRAY = '#c4c2d1'
    LAWN_GREEN = '#7dfc00'
    LIGHT_GRAY = '#d4d4d4'
    LIGHT_GREEN = '#8fed8f'
    LIGHT_YELLOW = '#ffffe0'
    LILAC = '#c7a3c7'
    LINEN = '#faf0e5'
    LISERAN_PURPLE = '#de70a1'

    MANATEE = '#9799ac'
    MAUVE = '#e1b1ff'
    MYSTIC = '#d75283'

    OPAL = '#a8c2bd'
    OUTRAGEOUS_ORANGE = '#ff6e4a'

    PEACH = '#ffe6b6'

    RASPBERRY = '#e40a5c'
    RUBY = '#e0125e'
    RUST = '#b8400d'

    SAND = '#c3b380'
    SAPPHIRE = '#0f52bb'
    SCARLET = '#ff2400'
    SEA_GREEN_CRAYOLA = '#00ffcc'
    SEASHELL = '#fff5ed'
    SUNGLOW = '#ffcd33'

    TAN = '#d2b68d'

    WHEAT = '#f6dfb3'

    ## Custom

    TERMINAL = '#0c0c0c'
    GRAY30 = '#1e1e1e'


class Colored:
    _win_init = False  # To make it work in Windows command prompt
    _RESET = '\033[0m'
    def __new__(cls, text: str, /, fg: str = Hex.MANATEE, bg: str = Hex.TERMINAL) -> str:
        """
        Return the colored version of `text`.

        ---

        ## Params
        - `fg`: foreground color in hexadecimal format
        - `bg`: background color in hexadecimal format
        """

        text = str(text)
        fg_r, fg_g, fg_b = hex_to_rgb(fg)
        bg_r, bg_g, bg_b = hex_to_rgb(bg)

        if _sys.platform.lower() == 'win32':
            if not Colored._win_init:
                _os.system('color')
                Colored._win_init = True

        header = (
            '\033['
            f'38;2;{fg_r};{fg_g};{fg_b};'
            f'48;2;{bg_r};{bg_g};{bg_b}m'
        )

        ## Handle multiple colors in one string
        if cls._RESET in text:
            text = text.replace(cls._RESET, header)

        return header + text + cls._RESET


from mykit.kit.color import (
    Hex as _Hex,
    Colored as _Colored
)
from mykit.kit.utils import printer as _printer


def _print_log(level, color, msg):
    _printer(_Colored(level, color) + ': ' + msg)


class pL:
    """pL (Print Log): A simple logger using Python's `print` function."""

    def debug(__msg: str, /) -> None:
        _print_log('DEBUG', _Hex.WHEAT, __msg)

    def info(__msg: str, /) -> None:
        _print_log('INFO', _Hex.BLUE_GRAY, __msg)

    def warning(__msg: str, /) -> None:
        _print_log('WARNING', _Hex.DARK_ORANGE, __msg)

    def error(__msg: str, /) -> None:
        _print_log('ERROR', _Hex.SCARLET, __msg)


import random as _random
import tkinter as _tk
from typing import (
    Dict as _Dict,
    List as _List,
    Optional as _Optional,
    Tuple as _Tuple,
    Union as _Union
)


class Plot:

    ## <runtime>

    _page: _tk.Canvas = None
    @staticmethod
    def _set_page(page):
        Plot._page = page

    plots: _Dict[str, 'Plot'] = {}
    plot_tags: _Dict[str, _List['Plot']] = {}

    ## </runtime>

    def __init__(
        self,
        points: _List[_Tuple[float, float]] = [],

        xmin: _Optional[float] = None,
        xmax: _Optional[float] = None,
        ymin: _Optional[float] = None,
        ymax: _Optional[float] = None,

        width: int = 300,
        height: int = 200,
        tl_x: int = 0,
        tl_y: int = 0,

        pad_x: float = 0.03,
        pad_y: float = 0.07,
        show_tick: bool = True,
        show_grid: bool = True,
        ntick_x: int = 10,
        ntick_y: int = 10,
        tick_len: int = 12,
        arrow_size: int = 7,
        arrow_width: int = 2,

        grid_color: str = '#555',
        axes_color: str = '#ccc',
        axes_label_color: str = '#ccc',

        title: str = '',
        title_color: str = '#fff',
        title_font: _Union[str, tuple] = ('Arial Bold', 15),

        x_axis_label: str = '',
        x_axis_label_shift: int = 15,
        x_axis_label_font: _Union[str, tuple] = ('Arial Bold', 12),
        y_axis_label: str = '',
        y_axis_label_shift: int = 15,
        y_axis_label_font: _Union[str, tuple] = ('Arial Bold', 12),

        tick_x_prefix: str = '',
        tick_x_suffix: str = '',
        tick_x_shift: int = 0,
        tick_x_font: _Union[str, tuple] = 'Consolas 9',
        tick_x_prec: int = 1,
        tick_y_prefix: str = '',
        tick_y_suffix: str = '',
        tick_y_shift: int = 0,
        tick_y_font: _Union[str, tuple] = 'Consolas 9',
        tick_y_prec: int = 1,
        tick_color: str = '#ccc',

        plot_color: str = '#7f7',
        plot_thick: int = 1,  # line width

        show_points: bool = False,
        points_rad: int = 7,
        points_color: str = '#a77',
        points_border: str = '#eee',

        visible: bool = True,

        id: _Optional[str] = None,
        tags: _Optional[_Union[str, _List[str]]] = None,
    ):
        """
        The graph can be displayed without any given points,
        but in order to show the plot, two points need to be specified.

        ---

        ## Params
        - `xrange`: if `None` -> using the x-range from `points`
        - `yrange`: if `None` -> using the y-range from `points`
        """

        ## <dependencies check>

        ## reminder: Plot isn't actually a widget. it can be used as a utility
        ##           outside App's purposes. so just use Plot._set_page to achieve it.
        if Plot._page is None:
            raise AssertionError('Can\'t use widgets before App initialized.')
        
        ## </dependencies check>

        self.points = points

        self.xmin = xmin
        self.xmax = xmax
        self.ymin = ymin
        self.ymax = ymax

        self.width = width
        self.height = height
        self.tl_x = tl_x
        self.tl_y = tl_y

        self.pad_x = pad_x
        self.pad_y = pad_y
        self.show_tick = show_tick
        self.show_grid = show_grid
        self.ntick_x = ntick_x
        self.ntick_y = ntick_y
        self.tick_len = tick_len
        self.arrow_size = arrow_size
        self.arrow_width = arrow_width

        self.grid_color = grid_color
        self.axes_color = axes_color
        self.axes_label_color = axes_label_color

        self.title = title
        self.title_color = title_color
        self.title_font = title_font

        self.x_axis_label = x_axis_label
        self.x_axis_label_shift = x_axis_label_shift
        self.x_axis_label_font = x_axis_label_font
        self.y_axis_label = y_axis_label
        self.y_axis_label_shift = y_axis_label_shift
        self.y_axis_label_font = y_axis_label_font

        self.tick_x_prefix = tick_x_prefix
        self.tick_x_suffix = tick_x_suffix
        self.tick_x_shift = tick_x_shift
        self.tick_x_font = tick_x_font
        self.tick_x_prec = tick_x_prec
        self.tick_y_prefix = tick_y_prefix
        self.tick_y_suffix = tick_y_suffix
        self.tick_y_shift = tick_y_shift
        self.tick_y_font = tick_y_font
        self.tick_y_prec = tick_y_prec
        self.tick_color = tick_color

        self.plot_color = plot_color
        self.plot_thick = plot_thick

        self.show_points = show_points
        self.points_rad = points_rad
        self.points_color = points_color
        self.points_border = points_border

        self.visibile = visible

        ## <id>

        ## `self.id`: to make sure that we can modify a specific instance without affecting the others
        if id is None:
            self.id = _random.randint(-10000, 10000)
            while self.id in Plot.plots:
                self.id = _random.randint(-10000, 10000)
        else:
            self.id = id
            if self.id in Plot.plots:
                raise ValueError(f'The id {repr(id)} is duplicated.')
        
        Plot.plots[self.id] = self

        ## </id>

        ## <tags>
        
        if type(tags) is str:
            self.tags = [tags]
        elif (type(tags) is list) or (type(tags) is tuple) or (tags is None):
            self.tags = tags

        if tags is not None:
            for tag in self.tags:
                if tag in Plot.plot_tags:
                    Plot.plot_tags[tag].append(self)
                else:
                    Plot.plot_tags[tag] = [self]
        
        ## </tags>


        ## <preprocessing>
        self.plot_width = width*(1-pad_x)
        self.plot_height = height*(1-pad_y)
        ## </preprocessing>


        ## init
        self._redraw()

    def _redraw(self):
        """redraw the entire graph"""

        ## 2 points need to be specified in order to draw the graph
        if len(self.points) < 3:
            self.points = [(0, 0), (1, 0)]

        x_values = [p[0] for p in self.points]
        if self.xmin is None:
            XMIN = min(x_values)
        else:
            XMIN = self.xmin
        if self.xmax is None:
            XMAX = max(x_values)
        else:
            XMAX = self.xmax

        y_values = [p[1] for p in self.points]
        if self.ymin is None:
            YMIN = min(y_values)
        else:
            YMIN = self.ymin
        if self.ymax is None:
            YMAX = max(y_values)
        else:
            YMAX = self.ymax

        LEN_X = XMAX - XMIN
        LEN_Y = YMAX - YMIN


        ## title
        Plot._page.create_text(
            self.tl_x+self.width/2, self.tl_y,
            text=self.title, font=self.title_font, fill=self.title_color,
            tags=f'Plot_{self.id}'
        )


        ## grids
        if self.show_grid:

            ## vertical grids
            for x in range(self.ntick_x):
                X = self.tl_x + ((x+1)/self.ntick_x)*self.plot_width
                Plot._page.create_line(
                    X, self.tl_y+(self.height-self.plot_height),
                    X, self.tl_y+self.height,
                    fill=self.grid_color, width=1, tags=f'Plot_{self.id}'
                )

            ## horizontal grids
            for y in range(self.ntick_y):
                Y = self.tl_y+self.height - ((y+1)/self.ntick_y)*self.plot_height
                Plot._page.create_line(
                    self.tl_x                , Y,
                    self.tl_x+self.plot_width, Y,
                    fill=self.grid_color, width=1, tags=f'Plot_{self.id}'
                )


        ## x-axis
        Plot._page.create_line(
            self.tl_x           , self.tl_y+self.height,
            self.tl_x+self.width, self.tl_y+self.height,
            fill=self.axes_color, width=1, tags=f'Plot_{self.id}'
        )
        ## x-axis arrow
        Plot._page.create_line(
            self.tl_x+self.width-self.arrow_size, self.tl_y+self.height-self.arrow_size,
            self.tl_x+self.width                , self.tl_y+self.height,
            self.tl_x+self.width-self.arrow_size, self.tl_y+self.height+self.arrow_size,
            fill=self.axes_color, width=self.arrow_width, tags=f'Plot_{self.id}'
        )
        ## x-axis label
        Plot._page.create_text(
            self.tl_x+self.width+self.x_axis_label_shift, self.tl_y+self.height,
            text=self.x_axis_label, anchor='w', fill=self.axes_label_color, font=self.x_axis_label_font, tags=f'Plot_{self.id}'
        )

        ## y-axis
        Plot._page.create_line(
            self.tl_x, self.tl_y,
            self.tl_x, self.tl_y+self.height,
            fill=self.axes_color, width=1, tags=f'Plot_{self.id}'
        )
        ## y-axis arrow
        Plot._page.create_line(
            self.tl_x-self.arrow_size, self.tl_y+self.arrow_size,
            self.tl_x, self.tl_y,
            self.tl_x+self.arrow_size, self.tl_y+self.arrow_size,
            fill=self.axes_color, width=self.arrow_width, tags=f'Plot_{self.id}'
        )
        ## y-axis label
        Plot._page.create_text(
            self.tl_x, self.tl_y-self.y_axis_label_shift,
            text=self.y_axis_label, anchor='s', fill=self.axes_label_color, font=self.y_axis_label_font, tags=f'Plot_{self.id}'
        )


        ## ticks
        if self.show_tick:

            ## x-axis ticks
            for x in range(self.ntick_x+1):
                X = self.tl_x + (x/self.ntick_x)*self.plot_width

                ## tick
                Plot._page.create_line(
                    X, self.tl_y+self.height-self.tick_len/2,
                    X, self.tl_y+self.height+self.tick_len/2,
                    fill=self.tick_color, width=1, tags=f'Plot_{self.id}'
                )

                ## tick-label
                _num = XMIN + (x/self.ntick_x)*(LEN_X)
                if self.tick_x_prec == 0:
                    _num = int(_num)
                else:
                    _num = round(_num, self.tick_x_prec)
                text = self.tick_x_prefix + str(_num) + self.tick_x_suffix
                Plot._page.create_text(
                    X, self.tl_y+self.height+self.tick_len+self.tick_x_shift,
                    text=text, anchor='n', font=self.tick_x_font, fill=self.axes_label_color,
                    tags=(f'Plot_{self.id}', f'Plot_{self.id}_ticks')
                )

            ## y-axis ticks
            for y in range(self.ntick_y+1):
                Y = self.tl_y+self.height - (y/self.ntick_y)*self.plot_height

                ## tick
                Plot._page.create_line(
                    self.tl_x-self.tick_len/2, Y,
                    self.tl_x+self.tick_len/2, Y,
                    fill=self.tick_color, width=1, tags=f'Plot_{self.id}'
                )

                ## tick-label
                _num = YMIN + (y/self.ntick_y)*(LEN_Y)
                if self.tick_y_prec == 0:
                    _num = int(_num)
                else:
                    _num = round(_num, self.tick_y_prec)
                text = self.tick_y_prefix + str(_num) + self.tick_y_suffix
                Plot._page.create_text(
                    self.tl_x-self.tick_len-self.tick_y_shift, Y,
                    text=text, anchor='e', font=self.tick_y_font, fill=self.axes_label_color,
                    tags=(f'Plot_{self.id}', f'Plot_{self.id}_ticks')
                )


        ## plot
        coords = []
        for x, y in self.points:
            X = self.tl_x + (x - XMIN)*(self.plot_width/LEN_X)
            Y = self.tl_y + self.height - (y - YMIN)*(self.plot_height/LEN_Y)
            coords.append((X, Y))
        Plot._page.create_line(
            coords,
            fill=self.plot_color, width=self.plot_thick,
            tags=(f'Plot_{self.id}', f'Plot_{self.id}_plot')
        )

        if self.show_points:
            for x, y in coords:
                Plot._page.create_oval(
                    x-self.points_rad/2, y-self.points_rad/2,
                    x+self.points_rad/2, y+self.points_rad/2,
                    fill=self.points_color, outline=self.points_border, width=1, tags=f'Plot_{self.id}'
                )
    
    def redraw_plot(self, points: _List[_Tuple[float, float]], /) -> None:
        """
        Redraws the plot and updates the tick labels with a new set of given `points`.
        """

        ## reminder: to optimize things, only redraw the necessary part
        ##           so the code below is duplicated from `_redraw`.
        ##           it's redundant, but currently the easiest way to
        ##           achieve the desired functionality

        self.points = points

        ## 2 points need to be specified in order to draw the graph
        if len(self.points) < 3:
            self.points = [(0, 0), (1, 0)]

        x_values = [p[0] for p in self.points]
        if self.xmin is None:
            XMIN = min(x_values)
        else:
            XMIN = self.xmin
        if self.xmax is None:
            XMAX = max(x_values)
        else:
            XMAX = self.xmax

        y_values = [p[1] for p in self.points]
        if self.ymin is None:
            YMIN = min(y_values)
        else:
            YMIN = self.ymin
        if self.ymax is None:
            YMAX = max(y_values)
        else:
            YMAX = self.ymax

        ## LEN_X and LEN_Y can't be zero
        LEN_X = max(10**(-self.tick_x_prec), XMAX - XMIN)
        LEN_Y = max(10**(-self.tick_y_prec), YMAX - YMIN)


        ## redraw the ticks

        Plot._page.delete(f'Plot_{self.id}_ticks')

        if self.show_tick:

            ## x-axis ticks
            for x in range(self.ntick_x+1):
                X = self.tl_x + (x/self.ntick_x)*self.plot_width

                ## tick-label
                _num = XMIN + (x/self.ntick_x)*(LEN_X)
                if self.tick_x_prec == 0:
                    _num = int(_num)
                else:
                    _num = round(_num, self.tick_x_prec)
                text = self.tick_x_prefix + str(_num) + self.tick_x_suffix
                Plot._page.create_text(
                    X, self.tl_y+self.height+self.tick_len+self.tick_x_shift,
                    text=text, anchor='n', font=self.tick_x_font, fill=self.axes_label_color,
                    tags=(f'Plot_{self.id}', f'Plot_{self.id}_ticks')
                )

            ## y-axis ticks
            for y in range(self.ntick_y+1):
                Y = self.tl_y+self.height - (y/self.ntick_y)*self.plot_height

                ## tick-label
                _num = YMIN + (y/self.ntick_y)*(LEN_Y)
                if self.tick_y_prec == 0:
                    _num = int(_num)
                else:
                    _num = round(_num, self.tick_y_prec)
                text = self.tick_y_prefix + str(_num) + self.tick_y_suffix
                Plot._page.create_text(
                    self.tl_x-self.tick_len-self.tick_y_shift, Y,
                    text=text, anchor='e', font=self.tick_y_font, fill=self.axes_label_color,
                    tags=(f'Plot_{self.id}', f'Plot_{self.id}_ticks')
                )


        ## redraw the plot

        Plot._page.delete(f'Plot_{self.id}_plot')

        coords = []
        for x, y in self.points:
            X = self.tl_x + (x - XMIN)*(self.plot_width/LEN_X)
            Y = self.tl_y + self.height - (y - YMIN)*(self.plot_height/LEN_Y)
            coords.append((X, Y))
        Plot._page.create_line(
            coords,
            fill=self.plot_color, width=self.plot_thick,
            tags=(f'Plot_{self.id}', f'Plot_{self.id}_plot')
        )

    def shift_plot(self, new_points: _List[_Tuple[float, float]], /) -> None:
        """
        Shifts the plot by inserting `new_points` and removing the leftmost points.
        For example, if 2 pairs are inserted, 2 leftmost pairs will be removed.
        Make sure that the new x values are always greater than the rightmost x value.
        For instance, if the current rightmost x is 100, the new x values
        should follow a sequence like 100, 101, 102.
        """
        n_new = len(new_points)
        points = self.points[n_new:] + new_points
        self.redraw_plot(points)

    def add_point(self, point: _Tuple[float, float], max: _Optional[int] = None):
        """
        Adding the new point to the current plot.
        If a max value is specified (e.g., max=100), and the total number of points
        exceeds 100, the plot will shift by inserting the new point and removing
        the leftmost point. If max is not specified, the point will simply be added.
        """

        points = self.points

        if max is None:
            points.append(point)
            self.redraw_plot(points)
        else:
            if len(points) < max:
                points.append(point)
                self.redraw_plot(points)
            else:
                self.shift_plot([point])

    
    def set_visibility(self, visible: bool, /):
        if self.visible is not visible:
            self.visible = visible
            self._redraw()

    @staticmethod
    def set_visibility_by_id(id: str, visible: bool, /):
        Plot.plots[id].set_visibility(visible)

    @staticmethod
    def set_visibility_by_tag(tag: str, visible: bool, /):
        for plot in Plot.plot_tags[tag]:
            plot.set_visibility(visible)

    @staticmethod
    def set_visibility_all(visible: bool, /):
        for plot in Plot.plots.values():
            plot.set_visibility(visible)


import random as _random
import tkinter as _tk
from typing import (
    Callable as _Callable,
    Dict as _Dict,
    List as _List,
    Literal as _Literal,
    Optional as _Optional,
    Tuple as _Tuple,
    Union as _Union
)


class Button:

    ## <runtime>

    _page: _tk.Canvas = None
    @staticmethod
    def _set_page(page):
        Button._page = page

    buttons: _Dict[str, 'Button'] = {}
    button_tags: _Dict[str, _List['Button']] = {}

    ## </runtime>

    def __init__(
        self,
        x: int = 0,
        y: int = 0,
        fn: _Optional[_Callable[[], None]] = None,
        label: str = '',
        label_font: _Union[str, _Tuple[str, int]] = 'Verdana 8',
        anchor: _Literal['center', 'n', 'ne', 'e', 'se', 's', 'sw', 'w', 'nw'] = 'nw',
        width: int = 100,
        height: int = 18,
        locked: bool = False,
        visible: bool = True,

        color_btn_normal: str = '#464646',
        color_btn_hover: str = '#5a5a5a',
        color_btn_press: str = '#6e6e6e',
        color_btn_locked: str = '#282828',
        color_bd_normal: str = '#6e6e6e',
        color_bd_locked: str = '#282828',
        color_lbl_normal: str = '#fafbfa',
        color_lbl_locked: str = '#050505',

        id: _Optional[str] = None,
        tags: _Optional[_Union[str, _List[str]]] = None,
    ) -> None:
        """
        ## Params
        - `x` and `y` is the position of the `anchor` (not the center of the button)
        - `color_btn_normal`: button's color
        - `color_bd_normal`: button's border color
        - `color_lbl_normal`: button's label color
        """

        ## <dependencies check>

        if Button._page is None:
            raise AssertionError('Can\'t use widgets before App initialized.')
        
        ## </dependencies check>

        self.x = x
        self.y = y
        self.fn = fn
        self.label = label
        self.label_font = label_font
        self.anchor = anchor
        self.width = width
        self.height = height
        self.locked = locked
        self.visible = visible

        self.color_btn_normal = color_btn_normal
        self.color_btn_hover = color_btn_hover
        self.color_btn_press = color_btn_press
        self.color_btn_locked = color_btn_locked
        self.color_bd_normal = color_bd_normal
        self.color_bd_locked = color_bd_locked
        self.color_lbl_normal = color_lbl_normal
        self.color_lbl_locked = color_lbl_locked

        ## <id>

        ## self.id ensures that we can modify a specific instance without affecting the others
        if id is None:
            self.id = str(_random.randint(0, 100_000))
            while self.id in Button.buttons:
                self.id = str(_random.randint(0, 100_000))
        else:
            self.id = id
            if self.id in Button.buttons:
                raise ValueError(f'The id {repr(id)} is duplicated.')
        
        Button.buttons[self.id] = self

        ## </id>

        ## <tags>

        if type(tags) is str:
            self.tags = [tags]
        elif (type(tags) is list) or (type(tags) is tuple) or (tags is None):
            self.tags = tags
        
        if tags is not None:
            for tag in self.tags:
                if tag in Button.button_tags:
                    Button.button_tags[tag].append(self)
                else:
                    Button.button_tags[tag] = [self]

        ## </tags>


        ## runtime

        self.default_label = label
        self.pressed = False
        self.hovered = False


        ## init

        self._redraw()


    def _redraw(self):

        if self.locked:
            color_btn = self.color_btn_locked
            color_bd = self.color_bd_locked
            color_lbl = self.color_lbl_locked
        elif self.pressed:
            color_btn = self.color_btn_press
            color_bd = self.color_bd_normal
            color_lbl = self.color_lbl_normal
        elif self.hovered:
            color_btn = self.color_btn_hover
            color_bd = self.color_bd_normal
            color_lbl = self.color_lbl_normal
        else:
            color_btn = self.color_btn_normal
            color_bd = self.color_bd_normal
            color_lbl = self.color_lbl_normal

        Button._page.delete(f'Button_{self.id}')

        if self.visible:

            ## This overhead will be executed each time this function is called.
            ## It may be inefficient, but it makes the code cleaner.
            X, Y = self.get_anchor_loc('center')  # the center of the button

            Button._page.create_rectangle(
                X - self.width/2, Y - self.height/2,
                X + self.width/2, Y + self.height/2,
                fill=color_btn, width=1, outline=color_bd,
                tags=f'Button_{self.id}'
            )
            Button._page.create_text(
                X, Y,
                text=self.label, font=self.label_font,
                fill=color_lbl,
                tags=f'Button_{self.id}'
            )


    def _hover(self):

        w2 = self.width/2
        h2 = self.height/2
        
        x = Button._page.winfo_pointerx()
        y = Button._page.winfo_pointery()

        X, Y = self.get_anchor_loc('center')

        ## `True` if the mouse cursor is inside the button
        inside = (X-w2 <= x <= X+w2) and (Y-h2 <= y <= Y+h2)

        if inside and (not self.locked) and self.visible and (not self.hovered):
            self.hovered = True
            self._redraw()  # just redraw once here

        elif self.hovered and (not inside):
            self.hovered = False
            self._redraw()  # just redraw once here

        ## reminder: don't put self._redraw below here because it will redraw regardless of the "hovered" state
        # self._redraw()

    @staticmethod
    def _hover_listener():  # reminder: hover listeners don't need `e: _tk.Event` arg
        for button in Button.buttons.values():
            button._hover()


    def _press(self):
        
        x = Button._page.winfo_pointerx()
        y = Button._page.winfo_pointery()

        X, Y = self.get_anchor_loc('center')
        w2 = self.width/2
        h2 = self.height/2
        inside = (X-w2 <= x <= X+w2) and (Y-h2 <= y <= Y+h2)

        if inside and (not self.locked) and self.visible:
            self.pressed = True
            self._redraw()

    @staticmethod
    def _press_listener(e: _tk.Event):
        for button in Button.buttons.values():
            button._press()


    def _release(self):
        if self.pressed:
            self.pressed = False
            self._redraw()            
            if self.fn is not None:
                self.fn()

    @staticmethod
    def _release_listener(e: _tk.Event):
        for button in list(Button.buttons.values()):
            button._release()


    def set_lock(self, locked: bool, /):
        if self.locked is not locked:
            self.locked = locked
            self._redraw()
    
    @staticmethod
    def set_lock_by_id(id: str, locked: bool, /):
        Button.buttons[id].set_lock(locked)

    @staticmethod
    def set_lock_by_tag(tag: str, locked: bool, /):
        for button in Button.button_tags[tag]:
            button.set_lock(locked)

    @staticmethod
    def set_lock_all(locked: bool, /):
        for button in Button.buttons.values():
            button.set_lock(locked)


    def set_visibility(self, visible: bool, /):
        if self.visible is not visible:
            self.visible = visible
            self._redraw()

    @staticmethod
    def set_visibility_by_id(id: str, visible: bool, /):
        Button.buttons[id].set_visibility(visible)

    @staticmethod
    def set_visibility_by_tag(tag: str, visible: bool, /):
        for button in Button.button_tags[tag]:
            button.set_visibility(visible)

    @staticmethod
    def set_visibility_all(visible: bool, /):
        for button in Button.buttons.values():
            button.set_visibility(visible)


    def set_label(self, label: _Optional[str], /):
        """
        If `label = None`, the default label (the one assigned
        when the instance was created) will be used.
        """

        if label is None:
            label = self.default_label

        if self.label != label:
            self.label = label
            self._redraw()

    @staticmethod
    def set_label_by_id(id: str, label: _Optional[str], /):
        Button.buttons[id].set_label(label)


    def set_fn(self, fn: _Callable[[], None], /):
        if self.fn is not fn:
            self.fn = fn

    @staticmethod
    def set_fn_by_id(id: str, fn: _Callable[[], None], /):
        Button.buttons[id].set_fn(fn)


    def get_anchor_loc(
        self,
        anchor: _Literal['center', 'n', 'ne', 'e', 'se', 's', 'sw', 'w', 'nw'],
        /
    ) -> _Tuple[int, int]:
        """To get the button's center coordinate, use `anchor='center'`"""

        W = self.width
        H = self.height

        ## get the coordinates for the center (X, Y)
        if self.anchor == 'center':
            X = self.x
            Y = self.y
        elif self.anchor == 'n':
            X = self.x
            Y = self.y + H/2
        elif self.anchor == 'ne':
            X = self.x - W/2
            Y = self.y + H/2
        elif self.anchor == 'e':
            X = self.x - W/2
            Y = self.y
        elif self.anchor == 'se':
            X = self.x - W/2
            Y = self.y - H/2
        elif self.anchor == 's':
            X = self.x
            Y = self.y - H/2
        elif self.anchor == 'sw':
            X = self.x + W/2
            Y = self.y - H/2
        elif self.anchor == 'w':
            X = self.x + W/2
            Y = self.y
        elif self.anchor == 'nw':
            X = self.x + W/2
            Y = self.y + H/2

        ## returning the requested anchor location
        if anchor == 'center':
            return (X, Y)
        elif anchor == 'n':
            return (X, Y-H/2)
        elif anchor == 'ne':
            return (X+W/2, Y-H/2)
        elif anchor == 'e':
            return (X+W/2, Y)
        elif anchor == 'se':
            return (X+W/2, Y+H/2)
        elif anchor == 's':
            return (X, Y+H/2)
        elif anchor == 'sw':
            return (X-W/2, Y+H/2)
        elif anchor == 'w':
            return (X-W/2, Y)
        elif anchor == 'nw':
            return (X-W/2, Y-H/2)
        else:
            raise ValueError(f'Invalid anchor value: {repr(anchor)}')

    @staticmethod
    def get_anchor_loc_by_id(
        id: str,
        anchor: _Literal['center', 'n', 'ne', 'e', 'se', 's', 'sw', 'w', 'nw'],
        /
    ) -> _Tuple[int, int]:
        return Button.buttons[id].get_anchor_loc(anchor)


    def move(
        self,
        x: int,
        y: int,
        /,
        anchor: _Optional[_Literal['center', 'n', 'ne', 'e', 'se', 's', 'sw', 'w', 'nw']] = None
    ) -> 'Button':
        """If `anchor = None`, the current anchor will be used."""
        self.x = x
        self.y = y
        if anchor is not None:
            self.anchor = anchor
        self._redraw()
        return self

    @staticmethod
    def move_by_id(
        id: str,
        x: int,
        y: int,
        /,
        anchor: _Optional[_Literal['center', 'n', 'ne', 'e', 'se', 's', 'sw', 'w', 'nw']] = None
    ) -> None:
        Button.buttons[id].move(x, y, anchor)


    def align(
        self,
        target: 'Button',
        anchor: str = 'nw',
        target_anchor: str = 'ne',
        xgap: float = 15,
        ygap: float = 0
    ) -> 'Button':
        """
        Valid options for `anchor` and `target_anchor` are `['center', 'n', 'ne', 'e', 'se', 's', 'sw', 'w', 'nw']`.
        """

        ## getting the target anchor location
        x, y = target.get_anchor_loc(target_anchor)

        ## shifting
        x += xgap
        y += ygap

        ## moving the label
        self.move(x, y, anchor)

        ## return the instance so that this method can also be used when
        ## creating the instance, like `btn_2 = Button().align(btn_1)`.
        return self


    def destroy(self) -> None:
        Button.buttons.pop(self.id)
        
        if self.tags is not None:
            for tag in self.tags:
                Button.button_tags[tag].remove(self)
                if Button.button_tags[tag] == []:
                    Button.button_tags.pop(tag)

        Button._page.delete(f'Button_{self.id}')
    
    @staticmethod
    def destroy_by_id(id: str, /) -> None:
        Button.buttons[id].destroy()

    @staticmethod
    def destroy_by_tag(tag: str, /) -> None:
        for button in list(Button.button_tags[tag]):
            button.destroy()

    @staticmethod
    def destroy_all() -> None:
        for button in list(Button.buttons.values()):
            button.destroy()



import math as _math
import re as _re
from typing import (
    NoReturn as _NoReturn,
    Union as _Union
)


def connum(num: _Union[str, int, float], /) -> _Union[str, _NoReturn]:
    """
    connum (Concise Number).
    Displaying the number as is, without plus signs or extra zeroes, for the UI.

    ---

    ## Exceptions
    - `ValueError` if `num` is not a number-like

    ## Demo
    - `connum(3.0)` -> `'3'`
    - `connum('+03.500')` -> `'3.5'`
    """
    res = _re.match(r'^(?P<sign>\+|-)?[0]*(?P<int>\d+)(?:(?P<dec>\.\d*?)[0]*)?$', str(num))
    if res is None:
        raise ValueError(f'Invalid numeric input: {repr(num)}.')

    sign, int, dec = res.groups()

    if sign in (None, '+'):
        ## Positive: omit the "+" sign (if any)
        out = ''
    else:
        if (int == '0') and (dec in (None, '.')):
            ## Handle "-0"
            out = ''
        else:
            out = '-'

    out += int

    if dec not in (None, '.'):
        ## Only if it's a decimal number (not like 123.0)
        out += dec

    return out


def byteFmt(__bytes: int, /) -> str:
    """in more readable byte format"""

    if __bytes == 0:
        return '0 B'

    exp = _math.floor(_math.log(abs(__bytes), 1024))
    val = round( __bytes / _math.pow(1024, exp), 2)

    UNIT = [
        'B', 'KiB', 'MiB',
        'GiB', 'TiB', 'PiB',
        'EiB', 'ZiB', 'YiB',
    ]

    return f'{val} {UNIT[exp]}'

def in_byte(bytes: int, /, precision: int = 2, gap: int = 1) -> str:
    """
    ## Params
    - `precision`: rounding precision
    - `gap`: gap (in spaces) between the number and the unit

    ## Demo
    >>> in_byte(100)  # 100 B
    >>> in_byte(1300)  # 1.27 KiB
    >>> in_byte(1300, 0, 0)  # 1KiB
    >>> in_byte(1700, 0, 0)  # 2KiB

    ## Docs
    - `in_byte` is the extended version of `mykit.kit.text.byteFmt`
    """

    GAP = ' '*gap

    ## Handle 0 `bytes`
    if bytes == 0: return '0' + GAP + 'B'

    ## Handle float `bytes`
    bytes = round(bytes)

    ## Handle negative `bytes`
    sign = ''
    if bytes < 0:
        bytes = abs(bytes)
        sign = '-'

    ## Calculate the exponent of 1024
    power = _math.floor( _math.log(bytes, 1024) )

    number = round( bytes / _math.pow(1024, power), precision )

    UNIT = [
        'B', 'KiB', 'MiB',
        'GiB', 'TiB', 'PiB',
        'EiB', 'ZiB', 'YiB',
    ]

    return sign + connum(number) + GAP + UNIT[power]


def num_approx(num, /, precision:int=1, gap:int=0) -> str:
    """
    Round a number down to K (thousand), M (million), B (billion), etc.

    ---

    ## Params
    - `precision`: rounding precision
    - `gap`: gap (in spaces) between the number and the unit

    ## Demo
    >>> num_approx(999)  # 999
    >>> num_approx(1000)  # 1K
    >>> num_approx(1001)  # 1K
    >>> num_approx(1_250_000)  # 1.2M
    >>> num_approx(1_250_000, 0, 1)  # 1 M
    >>> num_approx(1_750_000, 0, 1)  # 2 M
    """

    suffixes = [
        (1e3, ''),
        (1e6, 'K'),
        (1e9, 'M'),
        (1e12, 'B'),
        (1e15, 'T'),
        (1e18, 'q'),  # quadrillion
        (1e21, 'Q'),  # Quintillion
        (1e24, 's'),  # sextillion
        (1e27, 'S'),  # Septillion
    ]

    ## Handle negative `num`
    sign = ''
    if num < 0:
        num = abs(num)
        sign = '-'

    for divisor, suffix in suffixes:
        if num < divisor:
            n_format = f'{num*1000/divisor:.{precision}f}'
            break

    ## Remove trailing zeros after the decimal point if precision is greater than 0
    if precision > 0:
        n_format = n_format.rstrip('0').rstrip('.')
    
    GAP = ''
    if suffix != '':
        GAP = ' '*gap

    return sign + n_format + GAP + suffix


def num_round(number:int, /, keep:int=2, add_commas:bool=True) -> str:
    """
    Rounding number (see demo below).

    ---
    
    ## Params
    - `number`: The input number; floats are also accepted.
    - `keep`: number of the non-zero part (see demo below)
    - `add_commas`: separated the thousands with commas

    ## Demo
    >>> num_round(12345, 1)  # 10,000
    >>> num_round(12345, 2)  # 12,000
    >>> num_round(12345, 3)  # 12,300
    >>> num_round(12385, 3)  # 12,400

    ## Docs
    - `keep` value is clamped to a minimum of 1.
    """

    ## Handle float
    number = round(number)

    ## Handle negative number
    sign = ''
    if number < 0:
        number = abs(number)
        sign = '-'

    ## Clamp
    keep = min(len(str(number)), max(1, keep))
    
    ## Rounding
    power = len(str(number)) - keep
    scale = int( _math.pow(10, power) )
    rounded = round(number/scale)*scale

    ## Stringify
    if add_commas:
        rounded = f'{rounded:,}'
    else:
        rounded = str(rounded)

    return sign + rounded